import os
from typing import Optional

from aiohttp.web import Application

from .container import Database
from .dbmodel import CrudDB

__all__ = ["compile_query", "Database", "CrudDB", "get_db"]


def get_db(app: Application, store_url: Optional[str] = None) -> Optional[CrudDB]:
    """Create an Open API db handler and set it for use in an aiohttp application

    :param app: aiohttp Application
    :param store_url: datastore connection string, if not provided the env
        variable `DATASTORE` is used instead. If the env variable is not available
        either the method logs a warning and return `None`

    This function 1) adds the database to the aiohttp application at key "db",
    2) add the db command to the command line client (if command is True)
    and 3) add the close handler on application shutdown
    """
    store_url = store_url or os.environ.get("DATASTORE")
    if not store_url:  # pragma: no cover
        app.logger.warning("DATASTORE url not available")
        return None
    else:
        app["db"] = CrudDB(store_url)
        app.on_shutdown.append(close_db)
        return app["db"]


async def close_db(app: Application) -> None:
    await app["db"].close()



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union, cast

from sqlalchemy import Column, Table, func, select
from sqlalchemy.sql import Select, and_, or_
from sqlalchemy.sql.dml import Delete, Insert, Update

from ..db.container import Database
from ..pagination import (
    Pagination,
    PaginationVisitor,
    Search,
    SearchVisitor,
    fields_flip_sign,
)
from ..pagination.cursor import cursor_to_python
from ..types import Connection, Record, Records

QueryType = Union[Delete, Select, Update]
SelectUpdate = Union[Select, Update]


class CrudDB(Database):
    """A :class:`.Database` with additional methods for CRUD operations"""

    async def db_select(
        self,
        table: Table,
        filters: Dict,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> Records:
        """Select rows from a given table

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        sql_query = self.get_query(
            table, table.select(), consumer=consumer, params=filters
        )
        async with self.ensure_connection(conn) as conn:
            return await conn.execute(sql_query)

    async def db_delete(
        self,
        table: Table,
        filters: Dict,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> Records:
        """Delete rows from a given table

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        sql_query = self.get_query(
            table,
            table.delete().returning(*table.columns),
            consumer=consumer,
            params=filters,
        )
        async with self.ensure_connection(conn) as conn:
            return await conn.execute(sql_query)

    async def db_count(
        self,
        table: Table,
        filters: Dict,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> int:
        """Count rows in a table

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        sql_query = self.get_query(
            table, table.select(), consumer=consumer, params=filters
        )
        return await self.db_count_query(sql_query, conn=conn)

    async def db_count_query(
        self,
        sql_query,
        *,
        conn: Optional[Connection] = None,
    ) -> int:
        count_query = select([func.count()]).select_from(sql_query.alias("inner"))
        async with self.ensure_connection(conn) as conn:
            result = await conn.execute(count_query)
            return result.scalar()

    async def db_insert(
        self,
        table: Table,
        data: Union[List[Dict], Dict],
        *,
        conn: Optional[Connection] = None,
    ) -> Records:
        """Perform an insert into a table

        :param table: sqlalchemy Table
        :param data: key-value pairs for columns values
        :param conn: optional db connection
        """
        async with self.ensure_connection(conn) as conn:
            sql_query = self.insert_query(table, data)
            return await conn.execute(sql_query)

    async def db_update(
        self,
        table: Table,
        filters: Dict,
        data: Dict,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> Records:
        """Perform an update of rows

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows to update
        :param data: key-value pairs for updating columns values of selected rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        update = (
            cast(
                Update,
                self.get_query(
                    table, table.update(), consumer=consumer, params=filters
                ),
            )
            .values(**data)
            .returning(*table.columns)
        )
        async with self.ensure_connection(conn) as conn:
            return await conn.execute(update)

    async def db_upsert(
        self,
        table: Table,
        filters: Dict,
        data: Optional[Dict] = None,
        *,
        conn: Optional[Connection] = None,
        consumer: Any = None,
    ) -> Record:
        """Perform an upsert for a single record

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows to update
        :param data: key-value pairs for updating columns values of selected rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        if data:
            result = await self.db_update(
                table, filters, data, conn=conn, consumer=consumer
            )
        else:
            result = await self.db_select(table, filters, conn=conn, consumer=consumer)
        record = result.one_or_none()
        if record is None:
            insert_data = data.copy() if data else {}
            insert_data.update(filters)
            result = await self.db_insert(table, insert_data, conn=conn)
            record = result.one()
        return record

    async def db_paginate(
        self,
        table: Table,
        sql_query: Select,
        pagination: Pagination,
        *,
        conn: Optional[Connection] = None,
    ) -> Tuple[Records, Optional[int]]:
        pagination_visitor = DbPaginationVisitor(
            db=self, table=table, sql_query=sql_query
        )
        pagination.apply(pagination_visitor)
        async with self.ensure_connection(conn) as conn:
            return await pagination_visitor.execute(conn)

    # Query methods

    def insert_query(self, table: Table, records: Union[List[Dict], Dict]) -> Insert:
        if isinstance(records, dict):
            records = [records]
        else:
            cols: Set[str] = set()
            for record in records:
                cols.update(record)
            new_records = []
            for record in records:
                if len(record) < len(cols):
                    record = record.copy()
                    missing = cols.difference(record)
                    for col in missing:
                        record[col] = None
                new_records.append(record)
            records = new_records
        return table.insert(records).returning(*table.columns)

    # backward compatibility
    get_insert = insert_query

    def get_query(
        self,
        table: Table,
        sql_query: QueryType,
        *,
        params: Optional[Dict] = None,
        consumer: Any = None,
    ) -> QueryType:
        """Build an SqlAlchemy query

        :param table: sqlalchemy Table
        :param sql_query: sqlalchemy query type
        :param params: key-value pairs for the query
        :param consumer: optional consumer for manipulating parameters
        """
        filters: List = []
        columns = table.c
        params = params or {}

        for key, value in params.items():
            bits = key.split(":")
            field = bits[0]
            op = bits[1] if len(bits) == 2 else "eq"
            filter_field = getattr(consumer, f"filter_{field}", None)
            if filter_field:
                result = filter_field(op, value)
            else:
                field = getattr(columns, field)
                result = self.default_filter_field(field, op, value)
            if result is not None:
                if not isinstance(result, (list, tuple)):
                    result = (result,)
                filters.extend(result)
        if filters:
            whereclause = and_(*filters) if len(filters) > 1 else filters[0]
            sql_query = cast(Select, sql_query).where(whereclause)
        return sql_query

    def search_query(
        self, table: Table, sql_query: SelectUpdate, search: Search
    ) -> SelectUpdate:
        """Build an SqlAlchemy query for a search

        :param table: sqlalchemy Table
        :param sql_query: sqlalchemy query type
        :param search: the search dataclass
        """
        search_visitor = DbSearchVisitor(
            db=self, table=table, sql_query=cast(SelectUpdate, sql_query)
        )
        search.apply(search_visitor)
        return search_visitor.sql_query

    def order_by_query(
        self,
        table: Table,
        sql_query: Select,
        order_by: Optional[Union[str, Sequence[str]]],
    ) -> Select:
        """Apply ordering to a sql_query"""
        if isinstance(order_by, str):
            order_by = (order_by,)
        for name in order_by or ():
            if name.startswith("-"):
                order_by_column = getattr(table.c, name[1:], None)
                if order_by_column is not None:
                    order_by_column = order_by_column.desc()
            else:
                order_by_column = getattr(table.c, name, None)
            if order_by_column is not None:
                sql_query = sql_query.order_by(order_by_column)
        return sql_query

    # backward compatibility
    order_by = order_by_query

    def default_filter_field(self, field: Column, op: str, value: Any):
        """
        Applies a filter on a field.

        Notes on 'ne' op:

        Example data: [None, 'john', 'roger']
        ne:john would return only roger (i.e. nulls excluded)
        ne:     would return john and roger

        Notes on  'search' op:

        For some reason, SQLAlchemy uses to_tsquery rather than
        plainto_tsquery for the match operator

        to_tsquery uses operators (&, |, ! etc.) while
        plainto_tsquery tokenises the input string and uses AND between
        tokens, hence plainto_tsquery is what we want here

        For other database back ends, the behaviour of the match
        operator is completely different - see:
        http://docs.sqlalchemy.org/en/rel_1_0/core/sqlelement.html

        :param field: field name
        :param op: 'eq', 'ne', 'gt', 'lt', 'ge', 'le' or 'search'
        :param value: comparison value, string or list/tuple
        :return:
        """
        multiple = isinstance(value, (list, tuple))

        if multiple and op in ("eq", "ne"):
            if op == "eq":
                return field.in_(value)
            elif op == "ne":
                return ~field.in_(value)
        else:
            if multiple:
                assert len(value) > 0
                value = value[0]

            if op == "eq":
                return field == value
            elif op == "ne":
                return field != value
            elif op == "gt":
                return field > value
            elif op == "ge":
                return field >= value
            elif op == "lt":
                return field < value
            elif op == "le":
                return field <= value


@dataclass
class DbSearchVisitor(SearchVisitor):
    db: CrudDB
    table: Table
    sql_query: SelectUpdate

    def apply_search(self, search: str, search_fields: Sequence[str]) -> None:
        if search:
            columns = [getattr(self.table.c, col) for col in search_fields]
            self.sql_query = self.sql_query.where(
                or_(*(col.ilike(f"%{search}%") for col in columns))
            )


@dataclass
class DbPaginationVisitor(PaginationVisitor):
    db: CrudDB
    table: Table
    sql_query: Select
    initial_sql: Optional[QueryType] = None

    def apply_offset_pagination(
        self,
        limit: int,
        offset: int,
        order_by: Optional[Union[str, Sequence[str]]],
    ) -> None:
        self.initial_sql = self.sql_query
        sql_query = self.db.order_by_query(self.table, self.sql_query, order_by)
        if offset:
            sql_query = sql_query.offset(offset)
        if limit:
            sql_query = sql_query.limit(limit)
        self.sql_query = sql_query

    def apply_cursor_pagination(
        self,
        cursor: Sequence[Tuple[str, str]],
        limit: int,
        order_by: Sequence[str],
        previous: bool = False,
    ) -> None:
        sql_query = self.sql_query
        for key, value in cursor:
            sql_query = sql_query.where(self.filter(key, value, previous))
        extra = 1
        if previous:
            extra += 1
            order_by = fields_flip_sign(order_by)
        self.sql_query = self.db.order_by_query(self.table, sql_query, order_by).limit(
            limit + extra
        )

    async def execute(self, conn: Connection) -> Tuple[Records, Optional[int]]:
        total = None
        if self.initial_sql is not None:
            total = await self.db.db_count_query(self.initial_sql, conn=conn)
        values = await conn.execute(self.sql_query)
        return values, total

    def filter(self, field: str, value: str, previous: bool) -> Column:
        if field.startswith("-"):
            field = field[1:]
            op = "ge" if previous else "le"
        else:
            op = "le" if previous else "ge"
        column = getattr(self.table.c, field)
        py_value = cursor_to_python(column.type.python_type, value)
        return self.db.default_filter_field(column, op, py_value)



import asyncio
import logging
import re
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Sequence, Set

from .errors import ChannelCallbackError
from .utils import redis_to_py_pattern

CallbackType = Callable[[], None]

logger = logging.getLogger("trading.websocket")


@dataclass
class Event:
    name: str
    pattern: str
    regex: Any
    callbacks: Set[CallbackType] = field(default_factory=set)


@dataclass
class Channel:
    """A websocket channel"""

    name: str
    _events: Dict[str, Event] = field(default_factory=dict)

    @property
    def events(self):
        """List of event names this channel is registered with"""
        return tuple((e.name for e in self._events.values()))

    def __len__(self) -> int:
        return len(self._events)

    def __contains__(self, pattern: str) -> bool:
        return pattern in self._events

    def __iter__(self):
        return iter(self._events)

    def info(self) -> Dict:
        return {e.name: len(e.callbacks) for e in self._events.values()}

    async def __call__(self, message: Dict[str, Any]) -> Sequence[CallbackType]:
        """Execute callbacks from a new message

        Return callbacks which have raise WebsocketClosed or have raise an exception
        """
        event_name = message.get("event") or ""
        data = message.get("data")
        for event in tuple(self._events.values()):
            match = event.regex.match(event_name)
            if match:
                match = match.group()
                results = await asyncio.gather(
                    *[
                        self._execute_callback(callback, event, match, data)
                        for callback in event.callbacks
                    ]
                )
                return tuple(c for c in results if c)
        return ()

    def register(self, event_name: str, callback: CallbackType):
        """Register a ``callback`` for ``event_name``"""
        event_name = event_name or "*"
        pattern = self.event_pattern(event_name)
        event = self._events.get(pattern)
        if not event:
            event = Event(name=event_name, pattern=pattern, regex=re.compile(pattern))
            self._events[event.pattern] = event
        event.callbacks.add(callback)
        return event

    def get_subscribed(self, callback: CallbackType):
        events = []
        for event in self._events.values():
            if callback in event.callbacks:
                events.append(event.name)
        return events

    def unregister(self, event_name: str, callback: CallbackType):
        pattern = self.event_pattern(event_name)
        event = self._events.get(pattern)
        if event:
            return self.remove_event_callback(event, callback)

    def event_pattern(self, event):
        """Channel pattern for an event name"""
        return redis_to_py_pattern(event or "*")

    def remove_callback(self, callback: CallbackType) -> None:
        for key, event in tuple(self._events.items()):
            self.remove_event_callback(event, callback)

    def remove_event_callback(self, event: Event, callback: CallbackType) -> None:
        event.callbacks.discard(callback)
        if not event.callbacks:
            self._events.pop(event.pattern)

    async def _execute_callback(
        self, callback: CallbackType, event: Event, match: str, data: Any
    ):
        try:
            await callback(self.name, match, data)
        except ChannelCallbackError:
            return callback
        except Exception:
            logger.exception('callback exception: channel "%s" event "%s"', self, event)
            return callback



import os
from collections import OrderedDict
from dataclasses import MISSING, Field, asdict, dataclass, field
from dataclasses import fields as get_fields
from dataclasses import is_dataclass
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Set, Type, Union, cast

from aiohttp import hdrs, web

from ..data import fields
from ..data.exc import ErrorMessage, FieldError, ValidationErrors, error_response_schema
from ..exc import InvalidSpecException, InvalidTypeException
from ..utils import TypingInfo, compact, is_subclass
from .path import ApiPath
from .redoc import Redoc
from .server import default_server
from .utils import load_yaml_from_docstring, trim_docstring

OPENAPI = "3.0.3"
METHODS = [method.lower() for method in hdrs.METH_ALL]
SCHEMAS_TO_SCHEMA = ("response_schema", "body_schema")
SCHEMA_BASE_REF = "#/components/schemas/"

EMPTY_DEFAULTS = frozenset((None, MISSING, ""))
SPEC_ROUTE = os.environ.get("SPEC_ROUTE", "/spec")


@dataclass
class Contact:
    name: str = "API Support"
    url: str = "http://www.example.com/support"
    email: str = "support@example.com"


@dataclass
class License:
    name: str = "Apache 2.0"
    url: str = "https://www.apache.org/licenses/LICENSE-2.0.html"


@dataclass
class OpenApi:
    title: str = "Open API"
    description: str = ""
    version: str = "0.1.0"
    termsOfService: str = ""
    contact: Contact = Contact()
    license: License = License()


@dataclass
class OpenApiSpec:
    """Open API Specification"""

    info: OpenApi = field(default_factory=OpenApi)
    default_content_type: str = "application/json"
    default_responses: Dict = field(default_factory=dict)
    security: Dict = field(default_factory=dict)
    servers: List[Dict] = field(default_factory=list)
    validate_docs: bool = False
    allowed_tags: Set = field(default_factory=set)
    spec_url: str = SPEC_ROUTE
    redoc: Optional[Redoc] = None

    def routes(self, request: web.Request) -> Iterable:
        """Routes to include in the spec"""
        return request.app.router.routes()

    def setup_app(self, app: web.Application):
        app["spec"] = self
        app.router.add_get(self.spec_url, self.spec_route, name="openapi_spec")
        if self.redoc:
            app.router.add_get(self.redoc.path, self.redoc.handle_doc)

    def spec_route(self, request: web.Request) -> web.Response:
        """Return the OpenApi spec"""
        return web.json_response(self.build(request))

    def build(self, request: web.Request) -> Dict:
        doc = SpecDoc(request, self)
        security = self.security.copy()
        servers = self.servers[:] if self.servers else []
        return doc(security, servers)


class SchemaParser:
    """Utility class for parsing schemas"""

    def __init__(self, validate_docs: bool = False) -> None:
        self.validate_docs = validate_docs
        self.schemas_to_parse: Dict[str, type] = {}

    def get_parameters(self, schema: Any, default_in: str = "path") -> List:
        """Extract parameters list from a dataclass schema"""
        params = []
        json_schema = self.dataclass2json(schema)
        required = set(json_schema.get("required", ()))
        for name, entry in json_schema["properties"].items():
            entry = compact(
                name=name,
                description=entry.pop("description", None),
                schema=entry,
                required=name in required,
            )
            entry["in"] = default_in
            params.append(entry)
        return params

    def field2json(
        self, field_or_type: Union[Type, Field], validate: bool = True
    ) -> Dict[str, dict]:
        """Convert a dataclass field to Json schema"""
        field = fields.as_field(field_or_type)
        meta = field.metadata
        items = meta.get(fields.ITEMS)
        json_property = self.get_schema_info(field.type, items=items)
        field_description = meta.get(fields.DESCRIPTION)
        if not field_description:
            if self.validate_docs and validate:
                raise InvalidSpecException(
                    f'Missing description for field "{field.name}"'
                )
        else:
            json_property["description"] = field_description
        fmt = meta.get(fields.FORMAT)
        if fmt:
            json_property[fields.FORMAT] = fmt
        self.add_default(field, json_property)
        validator = meta.get(fields.VALIDATOR)
        # add additional parameters fields from validators
        if isinstance(validator, fields.Validator):
            validator.openapi(json_property)
        return json_property

    def dataclass2json(self, schema: Any) -> Dict[str, Any]:
        """Extract the object representation of a dataclass schema"""
        type_info = cast(TypingInfo, TypingInfo.get(schema))
        if not type_info or not type_info.is_dataclass:
            raise InvalidSpecException(
                "Schema must be a dataclass, got "
                f"{type_info.element if type_info else None}"
            )
        properties = {}
        required = []
        for item in get_fields(type_info.element):
            json_property = self.field2json(item)
            field_required = json_property.pop("required", True)
            if not json_property:
                continue
            if item.metadata.get(fields.REQUIRED, field_required):
                required.append(item.name)
            for name in fields.field_ops(item):
                properties[name] = json_property

        json_schema = {
            "type": "object",
            "description": trim_docstring(schema.__doc__ or ""),
            "properties": properties,
            "additionalProperties": False,
        }
        if required:
            json_schema["required"] = required
        return json_schema

    schema2json = dataclass2json
    # for backward compatibility

    def get_schema_info(
        self, schema: Any, items: Optional[Field] = None
    ) -> Dict[str, Any]:
        type_info = cast(TypingInfo, TypingInfo.get(schema))
        if type_info.container is list:
            return {
                "type": "array",
                "items": {"type": "object", "additionalProperties": True}
                if type_info.element is Any
                else self.field2json(
                    fields.as_field(type_info.element, field=items), False
                ),
            }
        elif type_info.container is dict:
            return {
                "type": "object",
                "additionalProperties": True
                if type_info.element is Any
                else self.field2json(
                    fields.as_field(type_info.element, field=items), False
                ),
            }
        elif type_info.is_union:
            required = True
            one_of = []
            for e in type_info.element:
                if e.is_none:
                    required = False
                else:
                    one_of.append(self.get_schema_info(e))
            info = one_of[0] if len(one_of) == 1 else {"oneOf": one_of}
            info["required"] = required
            return info
        elif type_info.is_dataclass:
            name = self.add_schema_to_parse(type_info.element)
            return {"$ref": f"{SCHEMA_BASE_REF}{name}"}
        else:
            return self.get_primitive_info(type_info.element)

    def get_primitive_info(self, schema: Type) -> Dict[str, Any]:
        mapping = fields.PRIMITIVE_TYPES.get(schema)
        if not mapping:
            if is_subclass(schema, Enum):
                enum_type = cast(Type[Enum], schema)
                return {"type": "string", "enum": [e.name for e in enum_type]}
            else:
                raise InvalidTypeException(f"Cannot understand {schema} while parsing")
        return dict(mapping)

    def add_schema_to_parse(self, schema: type) -> str:
        if not is_dataclass(schema):
            raise InvalidTypeException(f"Schema must be a dataclass, got {schema}")
        name = schema.__name__
        self.schemas_to_parse[name] = schema
        return name

    def parsed_schemas(self) -> Dict[str, Dict]:
        parsed = {}
        while self.schemas_to_parse:
            to_parse = self.schemas_to_parse
            self.schemas_to_parse = {}
            parsed.update(
                (
                    (name, self.dataclass2json(schema))
                    for name, schema in to_parse.items()
                )
            )
        return parsed

    def add_default(self, field: Field, json_property: Dict):
        if field.default in EMPTY_DEFAULTS or field.metadata.get(fields.REQUIRED):
            return
        default = field.default
        type_info = TypingInfo.get(field.type)
        if type_info.element not in fields.PRIMITIVE_TYPES:
            if is_subclass(type_info.element, Enum) and isinstance(
                default, type_info.element
            ):
                default = default.name
            else:
                return
        json_property["default"] = default


class SpecDoc(SchemaParser):
    """Build the OpenAPI Spec doc"""

    def __init__(
        self,
        request: web.Request,
        spec: OpenApiSpec,
        public: bool = True,
        private: bool = False,
    ) -> None:
        super().__init__(spec.validate_docs)
        self.request: web.Request = request
        self.spec: OpenApiSpec = spec
        self.logger = request.app.logger
        self.public: bool = public
        self.private: bool = private
        self.parameters: Dict = {}
        self.responses: Dict = {}
        self.tags: Dict = {}
        self.plugins: Dict = {}
        self.doc: Dict = dict(
            openapi=OPENAPI,
            info=asdict(self.spec.info or OpenApi()),
            paths=OrderedDict(),
        )

    @property
    def app(self) -> web.Application:
        return self.request.app

    def __call__(self, security: Dict, servers: List) -> Dict:
        # Add errors schemas
        self.add_schema_to_parse(ValidationErrors)
        self.add_schema_to_parse(ErrorMessage)
        self.add_schema_to_parse(FieldError)
        self.doc["security"] = [
            {name: value.pop("scopes", [])} for name, value in security.items()
        ]
        # Build paths
        self._build_paths()
        s = self.parsed_schemas()
        p = self.parameters
        r = self.responses
        doc = self.doc
        doc.update(
            compact(
                tags=[self.tags[name] for name in sorted(self.tags)],
                components=compact(
                    schemas=OrderedDict(((k, s[k]) for k in sorted(s))),
                    parameters=OrderedDict(((k, p[k]) for k in sorted(p))),
                    responses=OrderedDict(((k, r[k]) for k in sorted(r))),
                    securitySchemes=OrderedDict(
                        (((k, security[k]) for k in sorted(security)))
                    ),
                ),
                servers=servers,
            )
        )
        if not doc.get("servers"):
            # build the server info
            doc["servers"] = [default_server(self.request)]
        return doc

    # Internals

    def _build_paths(self) -> None:
        """Loop through app paths and add
        schemas, parameters and paths objects to the spec
        """
        paths = self.doc["paths"]
        base_path = self.app["cli"].base_path
        for route in self.spec.routes(self.request):
            route_info = route.get_info()
            path = route_info.get("path", route_info.get("formatter", None))
            handler = route.handler
            if is_subclass(handler, ApiPath) and self._include(handler.private):
                N = len(base_path)
                path = path[N:]
                paths[path] = self._build_path_object(path, handler)

        if self.validate_docs:
            self._validate_tags()

    def _validate_tags(self) -> None:
        for tag_name, tag_obj in self.tags.items():
            if self.spec.allowed_tags and tag_name not in self.spec.allowed_tags:
                raise InvalidSpecException(f'Tag "{tag_name}" not allowed')
            if "description" not in tag_obj:
                raise InvalidSpecException(f'Missing tag "{tag_name}" description')

    def _build_path_object(self, path: str, handler):
        path_obj = load_yaml_from_docstring(handler.__doc__) or {}
        doc_tags = path_obj.pop("tags", None)
        if not doc_tags and self.validate_docs:
            raise InvalidSpecException(f"Missing tags docstring for route '{path}'")

        tags = self._extend_tags(doc_tags)
        if handler.path_schema:
            path_obj["parameters"] = self.get_parameters(handler.path_schema)
        for method in METHODS:
            try:
                method_handler = getattr(handler, method, None)
                if method_handler is None:
                    continue

                operation = getattr(method_handler, "op", None)
                if operation is None:
                    self.logger.warning(
                        "No operation defined for %s.%s", handler.__name__, method
                    )
                    continue

                method_doc = load_yaml_from_docstring(method_handler.__doc__) or {}
                if not self._include(method_doc.pop("private", self.private)):
                    continue
                mtags = tags.copy()
                mtags.update(self._extend_tags(method_doc.pop("tags", None)))
                self._get_response_object(operation.response_schema, method_doc)
                self._get_request_body_object(operation.body_schema, method_doc)
                self._get_query_parameters(operation.query_schema, method_doc)
                method_info = self._get_method_info(method_handler, method_doc)
                method_doc.update(method_info)
                method_doc["tags"] = list(mtags)
                path_obj[method] = method_doc
            except (InvalidSpecException, InvalidTypeException) as exc:
                raise InvalidSpecException(
                    f"Invalid spec in route '{method} {path}': {exc}"
                ) from None

        return path_obj

    def _get_method_info(self, method_handler, method_doc):
        summary = method_doc.get("summary", "")
        description = method_doc.get("description", "")
        if self.validate_docs:
            if not summary:
                raise InvalidSpecException(
                    f'Missing method summary for "{method_handler}"'
                )
            if not description:
                raise InvalidSpecException(
                    f'Missing method description for "{method_handler}"'
                )
        return {"summary": summary, "description": description}

    def _get_response_object(
        self, type_info: Optional[TypingInfo], doc: Dict[str, str]
    ) -> None:
        if type_info:
            schema = self.get_schema_info(type_info)
            responses = {}
            for response, data in doc.get("responses", {}).items():
                rschema = schema
                if response >= 400:
                    rschema = self.get_schema_info(error_response_schema(response))
                content = data.get("content", self.spec.default_content_type)
                responses[response] = {
                    "description": data.get("description", ""),
                    "content": {content: {"schema": rschema}},
                }
            doc["responses"] = responses

    def _get_request_body_object(
        self, type_info: Optional[TypingInfo], doc: Dict[str, str]
    ) -> None:
        if type_info:
            content = doc.pop("body_content", self.spec.default_content_type)
            schema = self.get_schema_info(type_info)
            doc["requestBody"] = {"content": {content: {"schema": schema}}}

    def _get_query_parameters(
        self, type_info: Optional[TypingInfo], doc: Dict[str, str]
    ) -> None:
        if type_info:
            doc["parameters"] = self.get_parameters(type_info, "query")

    def _extend_tags(self, tags):
        names = set()
        for tag in tags or ():
            if isinstance(tag, str):
                tag = {"name": tag}
            name = tag.get("name")
            if name:
                if name not in self.tags:
                    self.tags[name] = tag
                else:
                    self.tags[name].update(tag)
                names.add(name)
        return names

    def _include(self, is_private: bool) -> bool:
        return (is_private and self.private) or (not is_private and self.public)



from dataclasses import dataclass

from aiohttp import web


@dataclass
class Redoc:
    """A dataclass for redoc rendering"""

    path: str = "/docs"
    favicon_url: str = (
        "https://raw.githubusercontent.com/Redocly/redoc/master/demo/favicon.png"
    )
    redoc_js_url: str = (
        "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"
    )
    font: str = "family=Montserrat:300,400,700|Roboto:300,400,700"

    def handle_doc(self, request: web.Request) -> web.Response:
        """Render a webpage with redoc and the spec form the app"""
        spec = request.app["spec"]
        spec_url = request.app.router["openapi_spec"].url_for()
        title = spec.info.title
        html = f"""
        <!DOCTYPE html>
        <html>
        <head>
        <title>{title}</title>
        <!-- needed for adaptive design -->
        <meta charset="utf-8"/>
        <meta name="viewport" content="width=device-width, initial-scale=1">
        """
        if self.font:
            html += f"""
        <link href="https://fonts.googleapis.com/css?{self.font}" rel="stylesheet">
        """
        html += f"""
        <link rel="shortcut icon" href="{self.favicon_url}">
        <!--
        ReDoc doesn't change outer page styles
        -->
        <style>
        body {{
            margin: 0;
            padding: 0;
        }}
        </style>
        </head>
        <body>
        <redoc spec-url="{spec_url}"></redoc>
        <script src="{self.redoc_js_url}"> </script>
        </body>
        </html>
        """
        return web.Response(text=html, content_type="text/html")



from .operation import op
from .path import ApiPath
from .redoc import Redoc
from .spec import OpenApi, OpenApiSpec, SchemaParser, SpecDoc

__all__ = [
    "op",
    "ApiPath",
    "OpenApi",
    "OpenApiSpec",
    "SchemaParser",
    "SpecDoc",
    "Redoc",
]



"""Testing utilities
"""
import asyncio
from contextlib import asynccontextmanager, contextmanager
from typing import Any

from aiohttp.client import ClientResponse
from aiohttp.test_utils import TestClient, TestServer
from aiohttp.web import Application

from .db import CrudDB, Database
from .json import dumps, loads
from .types import Connection


async def json_body(response: ClientResponse, status: int = 200) -> Any:
    assert response.content_type == "application/json"
    data = await response.json(loads=loads)
    if response.status != status:  # pragma: no cover
        print(dumps({"status": response.status, "data": data}, indent=4))

    assert response.status == status
    return data


@contextmanager
def with_test_db(db: CrudDB) -> CrudDB:
    db.create_all()
    try:
        yield db
    finally:
        db.drop_all_schemas()


class SingleConnDatabase(CrudDB):  # noqa
    """Useful for speedup testing"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._lock = asyncio.Lock()
        self._connection = None

    @classmethod
    def from_db(cls, db: Database) -> "SingleConnDatabase":
        return cls(dsn=db.dsn, metadata=db.metadata)

    async def __aenter__(self) -> "SingleConnDatabase":
        self._connection = await self.engine.begin()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        transaction = self._connection.get_transaction()
        await transaction.rollback()
        self._connection = None

    @asynccontextmanager
    async def connection(self) -> Connection:
        async with self._lock:
            yield self._connection

    @asynccontextmanager
    async def transaction(self) -> Connection:
        async with self._lock:
            yield self._connection


@asynccontextmanager
async def app_cli(app: Application) -> TestClient:
    server = TestServer(app)
    client = TestClient(server, json_serialize=dumps)
    await client.start_server()
    try:
        yield client
    finally:
        await client.close()



from dataclasses import dataclass
from functools import wraps
from typing import Any

from ..data.validate import ValidationErrors, validate


@dataclass
class ws_rpc:
    """Defines a Websocket RPC method in an OpenAPI Path"""

    body_schema: Any = None
    response_schema: Any = None

    def __call__(self, method):
        method.ws_rpc = self

        @wraps(method)
        async def _(view, payload):
            if self.body_schema:
                d = validate(self.body_schema, payload)
                if d.errors:
                    raise ValidationErrors(d.errors)
                payload = d.data
            data = await method(view, payload)
            return view.dump(self.response_schema, data)

        return _



from dataclasses import asdict, fields
from typing import Any, Dict, List, Optional, Union, cast

from openapi.types import Record

from ..utils import TypingInfo, iter_items
from .fields import DUMP


def is_nothing(value: Any) -> bool:
    if value == 0 or value is False:
        return False
    return not value


def dump(schema: Any, data: Any) -> Any:
    """Dump data with a given schema.

    :param schema: a valid :ref:`aio-openapi-schema`
    :param data: data to dump, if dataclasses are part of the schema,
        the `dump` metadata function will be used if available (see :func:`.data_field`)
    """
    type_info = cast(TypingInfo, TypingInfo.get(schema))
    if type_info.container is list:
        return dump_list(type_info.element, cast(List, data))
    elif type_info.container is dict:
        return dump_dict(type_info.element, cast(Dict, data))
    elif type_info.is_dataclass:
        return dump_dataclass(type_info.element, data)
    else:
        return data


def dump_dataclass(schema: Any, data: Optional[Union[Dict, Record]] = None) -> Dict:
    """Dump a dictionary of data with a given dataclass dump functions
    If the data is not given, the schema object is assumed to be
    an instance of a dataclass.
    """
    if data is None:
        data = asdict(schema)
    elif isinstance(data, schema):
        data = asdict(data)
    cleaned = {}
    fields_ = {f.name: f for f in fields(schema)}
    for name, value in iter_items(data):
        if name not in fields_ or is_nothing(value):
            continue
        field = fields_[name]
        dump_value = field.metadata.get(DUMP)
        if dump_value:
            value = dump_value(value)
        cleaned[field.name] = dump(field.type, value)

    return cleaned


def dump_list(schema: Any, data: List) -> List[Dict]:
    """Validate a dictionary of data with a given dataclass"""
    return [dump(schema, d) for d in data]


def dump_dict(schema: Any, data: Dict[str, Any]) -> List[Dict]:
    """Validate a dictionary of data with a given dataclass"""
    return {name: dump(schema, d) for name, d in data.items()}



from dataclasses import dataclass
from typing import List

from .fields import data_field, str_field


@dataclass
class ErrorMessage:
    """Error message and list of errors for data fields"""

    message: str = str_field(description="Error message")


@dataclass
class FieldError(ErrorMessage):
    """Error message for a data field"""

    field: str = str_field(description="name of the data field with error")


@dataclass
class ValidationErrors(ErrorMessage):
    """Error message and list of errors for data fields"""

    errors: List[FieldError] = data_field(description="List of field errors")


def error_response_schema(status):
    return ValidationErrors if status == 422 else ErrorMessage



import logging
import os

import click

try:
    import colorlog
except ImportError:  # pragma: no cover
    colorlog = None


LEVEL = (os.environ.get("LOG_LEVEL") or "info").upper()
LOGGER_NAME = os.environ.get("APP_NAME") or ""
LOG_FORMAT = "%(levelname)s: %(name)s: %(message)s"

logger = logging.getLogger(LOGGER_NAME)


def get_logger(name: str = "") -> logging.Logger:
    return logger.getChild(name) if name else logger


getLogger = get_logger


@click.pass_context
def setup_logging(ctx, verbose, quiet):
    if verbose:
        level = "DEBUG"
    elif quiet:
        level = "ERROR"
    else:
        level = LEVEL
    level = getattr(logging, level) if level != "NONE" else None
    ctx.obj["log_level"] = level
    if level:
        logger.setLevel(level)
        if not logger.hasHandlers():
            fmt = LOG_FORMAT
            if colorlog:
                handler = colorlog.StreamHandler()
                fmt = colorlog.ColoredFormatter(f"%(log_color)s{LOG_FORMAT}")
            else:  # pragma: no cover
                handler = logging.StreamHandler()
                fmt = logging.Formatter(LOG_FORMAT)
            handler.setFormatter(fmt)
            logger.addHandler(handler)



from typing import Any, Dict, List, Type, Union

from multidict import MultiDict
from sqlalchemy.engine import CursorResult, Row
from sqlalchemy.ext.asyncio import AsyncConnection

PrimitiveType = Union[int, float, bool, str]
JSONType = Union[PrimitiveType, List, Dict[str, Any]]
DataType = Any

SchemaType = Union[List[Type], Type]
SchemaTypeOrStr = Union[str, SchemaType]
StrDict = Dict[str, Any]
QueryType = Union[StrDict, MultiDict]
Record = Row
Records = CursorResult
Connection = AsyncConnection



from typing import Callable, Dict, List, Optional, Sequence

from aiohttp.web import Application

from .cli import OpenApiClient
from .spec import OpenApi, OpenApiSpec, Redoc


def rest(
    openapi: Optional[Dict] = None,
    setup_app: Callable[[Application], None] = None,
    base_path: str = "",
    commands: Optional[List] = None,
    allowed_tags: Sequence[str] = (),
    validate_docs: bool = False,
    servers: Optional[List[str]] = None,
    security: Optional[Dict[str, Dict]] = None,
    redoc: Optional[Redoc] = None,
    OpenApiSpecClass: type = OpenApiSpec,
    **kwargs,
) -> OpenApiClient:
    """Create the OpenApi Client"""
    if openapi is not None:
        openapi = OpenApiSpecClass(
            OpenApi(**(openapi or {})),
            allowed_tags=allowed_tags,
            validate_docs=validate_docs,
            servers=servers,
            security=security,
            redoc=redoc,
        )
    return OpenApiClient(
        spec=openapi,
        base_path=base_path,
        commands=commands,
        setup_app=setup_app,
        **kwargs,
    )



import enum
from dataclasses import Field, dataclass, field, fields
from datetime import date, datetime, time
from decimal import Decimal, InvalidOperation
from numbers import Number
from typing import Any, Callable, Dict, Iterator, Optional, Tuple
from uuid import UUID

from dateutil.parser import parse as parse_date
from email_validator import EmailNotValidError, validate_email

from .. import json, tz
from ..utils import compact_dict, str2bool

REQUIRED = "required"
VALIDATOR = "OPENAPI_VALIDATOR"
DESCRIPTION = "description"
POST_PROCESS = "post_process"
DUMP = "dump"
FORMAT = "format"
OPS = "ops"
ITEMS = "items"


PRIMITIVE_TYPES: Dict[Any, Dict] = {
    str: {"type": "string"},
    bytes: {"type": "string", FORMAT: "binary"},
    int: {"type": "integer", FORMAT: "int32"},
    float: {"type": "number", FORMAT: "float"},
    bool: {"type": "boolean"},
    date: {"type": "string", FORMAT: "date"},
    datetime: {"type": "string", FORMAT: "date-time"},
    Decimal: {"type": "number"},
}


class Ops(enum.Enum):
    eq = enum.auto()
    ne = enum.auto()
    gt = enum.auto()
    ge = enum.auto()
    lt = enum.auto()
    le = enum.auto()


class ValidationError(ValueError):
    def __init__(self, field: str, message: str) -> None:
        self.field = field
        self.message = message


def field_dict(dc: type) -> Dict[str, Field]:
    return {f.name: f for f in fields(dc)}


def data_field(
    required: bool = False,
    validator: Optional[Callable[[Field, Any], Any]] = None,
    dump: Optional[Callable[[Any], Any]] = None,
    format: str = None,
    description: str = None,
    items: Optional[Field] = None,
    post_process: Callable[[Any], Any] = None,
    ops: Tuple = (),
    meta: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Field:
    """Extend a dataclass field with the following metadata

    :param validator: optional callable which accept field and raw value
        as inputs and return the validated value
    :param required: boolean specifying if field is required
    :param dump: optional callable which receive the field value and convert to
        the desired value to serve in requests
    :param format: optional string which represents the JSON schema format
    :param description: optional field description
    :param items: field for items of the current field
        (only used for `List` and `Dict` fields)
    :param post_process: post processor function executed after validation
    :param ops: optional tuple of strings specifying available operations
    :param meta: optional dictionary with additional metadata
    """
    if isinstance(validator, Validator) and not dump:
        dump = validator.dump
    # Add default None otherwisenon-default fields can follow default ones
    if "default_factory" not in kwargs:
        kwargs.setdefault("default", None)
    meta = meta or {}

    f = field(
        metadata=compact_dict(
            {
                VALIDATOR: validator,
                REQUIRED: required,
                DUMP: dump,
                DESCRIPTION: description,
                ITEMS: items,
                POST_PROCESS: post_process,
                FORMAT: format,
                OPS: ops,
                **meta,
            }
        ),
        **kwargs,
    )
    return f


def str_field(min_length: int = 0, max_length: int = 0, **kw) -> Field:
    """A specialized :func:`.data_field` for strings

    :param min_length: minim length of string
    :param max_length: maximum length of string
    """
    kw.setdefault(
        "validator", StrValidator(min_length=min_length, max_length=max_length)
    )
    return data_field(**kw)


def bool_field(**kw) -> Field:
    """Specialized :func:`.data_field` for bool types"""
    kw.setdefault("validator", BoolValidator())
    return data_field(**kw)


def uuid_field(format: str = "uuid", **kw) -> Field:
    """A UUID field with validation"""
    kw.setdefault("validator", UUIDValidator())
    return data_field(format=format, **kw)


def number_field(
    min_value: Optional[Number] = None,
    max_value: Optional[Number] = None,
    precision: Optional[int] = None,
    **kw,
) -> Field:
    """A specialized :func:`.data_field` for numeric values

    :param min_value: minimum value
    :param max_value: maximum value
    :param precision: decimal precision
    """
    kw.setdefault("validator", NumberValidator(min_value, max_value, precision))
    return data_field(**kw)


def integer_field(
    min_value: Optional[Number] = None,
    max_value: Optional[Number] = None,
    **kw,
) -> Field:
    """A specialized :func:`.data_field` for integer values

    :param min_value: minimum value
    :param max_value: maximum value
    """
    kw.setdefault("validator", IntegerValidator(min_value, max_value))
    return data_field(**kw)


def decimal_field(min_value=None, max_value=None, precision=None, **kw) -> Field:
    kw.setdefault("validator", DecimalValidator(min_value, max_value, precision))
    return data_field(**kw)


def email_field(min_length: int = 0, max_length: int = 0, **kw) -> Field:
    """A specialized :func:`.data_field` for emails, validation via the
    `email_validator` third party library

    :param min_length: minimum length of email
    :param max_length: maximum length of email
    """
    kw.setdefault(
        "validator", EmailValidator(min_length=min_length, max_length=max_length)
    )
    return data_field(**kw)


def enum_field(EnumClass, **kw) -> Field:
    """A specialized :func:`.data_field` for enums

    :param EnumClass: enum for validation
    """
    kw.setdefault("validator", EnumValidator(EnumClass))
    return data_field(**kw)


def date_field(**kw) -> Field:
    """A specialized :func:`.data_field` for dates"""
    kw.setdefault("validator", DateValidator())
    return data_field(**kw)


def date_time_field(timezone=False, **kw) -> Field:
    """A specialized :func:`.data_field` for datetimes

    :param timezone: timezone for validation
    """
    kw.setdefault("validator", DateTimeValidator(timezone=timezone))
    return data_field(**kw)


def as_field(item: Any, *, field: Optional[Field] = None, **kw) -> Field:
    if isinstance(item, Field):
        return item
    field = field or data_field(**kw)
    if field.type and field.type is not item:
        raise RuntimeError("Cannot override field type")
    field.type = item
    return field


def json_field(**kw) -> Field:
    """A specialized :func:`.data_field` for JSON data"""
    kw.setdefault("validator", JSONValidator())
    return data_field(**kw)


# Utilities


def field_ops(field: Field) -> Iterator[str]:
    yield field.name
    for op in field.metadata.get(OPS, ()):
        yield f"{field.name}:{op}"


# VALIDATORS


class Validator:
    def __call__(self, field: Field, value: Any) -> Any:
        raise ValidationError(field.name, "invalid")

    def openapi(self, prop: Dict) -> None:
        pass

    def dump(self, value: Any) -> Any:
        return value


@dataclass
class StrValidator(Validator):
    min_length: int = 0
    max_length: int = 0

    def __call__(self, field: Field, value: Any) -> Any:
        if not isinstance(value, str):
            raise ValidationError(field.name, "Must be a string")
        if self.min_length and len(value) < self.min_length:
            raise ValidationError(field.name, "Too short")
        if self.max_length and len(value) > self.max_length:
            raise ValidationError(field.name, "Too long")
        return value

    def openapi(self, prop: Dict) -> None:
        if self.min_length:
            prop["minLength"] = self.min_length
        if self.max_length:
            prop["maxLength"] = self.max_length


@dataclass
class EmailValidator(StrValidator):
    def __call__(self, field: Field, value: Any) -> Any:
        value = super().__call__(field, value)
        try:
            validate_email(value, check_deliverability=False)
        except EmailNotValidError:
            raise ValidationError(field.name, "%s not a valid email" % value) from None
        return value


class ListValidator(Validator):
    def __init__(self, validators) -> None:
        self.validators = validators

    def __call__(self, field: Field, value: Any) -> Any:
        for validator in self.validators:
            value = validator(field, value)
        return value

    def dump(self, value: Any) -> Any:
        for validator in self.validators:
            dump = getattr(validator, "dump", None)
            if hasattr(dump, "__call__"):
                value = dump(value)
        return value

    def openapi(self, prop: Dict) -> None:
        for validator in self.validators:
            if isinstance(validator, Validator):
                validator.openapi(prop)


class UUIDValidator(Validator):
    def __call__(self, field: Field, value: Any) -> Any:
        try:
            if not isinstance(value, UUID):
                value = UUID(str(value))
            return value.hex
        except ValueError:
            raise ValidationError(field.name, "%s not a valid uuid" % value)

    def dump(self, value: Any) -> Any:
        if isinstance(value, UUID):
            return value.hex
        return value


class EnumValidator(Validator):
    """Enum validator to and from name (str) and value (int)"""

    def __init__(self, EnumClass) -> None:
        self.EnumClass = EnumClass

    def __call__(self, field: Field, value: Any) -> Any:
        try:
            e = value
            if isinstance(e, str):
                e = getattr(self.EnumClass, value)
            if isinstance(e, self.EnumClass):
                return e if field.type == self.EnumClass else e.name
            raise AttributeError
        except AttributeError:
            raise ValidationError(field.name, "%s not valid" % value)

    def dump(self, value: Any) -> Any:
        if isinstance(value, self.EnumClass):
            return value.name
        return value


class Choice(Validator):
    def __init__(self, choices) -> None:
        self.choices = choices

    def __call__(self, field: Field, value: Any) -> Any:
        if value not in self.choices:
            raise ValidationError(field.name, "%s not valid" % value)
        return value

    def openapi(self, prop: Dict) -> None:
        prop["enum"] = sorted(self.choices)


class DateValidator(Validator):
    def dump(self, value: Any) -> Any:
        if isinstance(value, datetime):
            return value.date().isoformat()
        elif isinstance(value, date):
            return value.isoformat()
        return value

    def __call__(self, field: Field, value: Any) -> Any:
        if isinstance(value, str):
            try:
                value = parse_date(value).date()
            except ValueError:
                pass
        if not isinstance(value, date):
            raise ValidationError(field.name, "%s not valid format" % value)
        return value


class DateTimeValidator(Validator):
    def __init__(self, timezone=False) -> None:
        self.timezone = timezone

    def dump(self, value: Any) -> Any:
        if isinstance(value, datetime):
            return value.isoformat()
        return value

    def __call__(self, field: Field, value: Any) -> Any:
        if isinstance(value, str):
            try:
                value = parse_date(value)
            except ValueError:
                pass
        if not isinstance(value, datetime):
            raise ValidationError(field.name, "%s not valid format" % value)
        if self.timezone and not value.tzinfo:
            if value.time() == time():
                value = tz.as_utc(value)
            else:
                raise ValidationError(field.name, "Timezone information required")
        return value


NumericErrors = (TypeError, ValueError, InvalidOperation)


class BoundedNumberValidator(Validator):
    def __init__(
        self, min_value: Optional[Number] = None, max_value: Optional[Number] = None
    ) -> None:
        self.min_value = min_value
        self.max_value = max_value

    def __call__(self, field: Field, value: Any) -> Any:
        if self.min_value is not None and value < self.min_value:
            raise ValidationError(
                field.name, "%s less than %s" % (value, self.min_value)
            )
        if self.max_value is not None and value > self.max_value:
            raise ValidationError(
                field.name, "%s greater than %s" % (value, self.max_value)
            )
        return value

    def dump(self, value: Any) -> Number:
        return self.to_number(value)

    def openapi(self, prop: Dict) -> None:
        if self.min_value is not None:
            prop["minimum"] = self.min_value
        if self.max_value is not None:
            prop["maximum"] = self.max_value

    def to_number(self, value: Any) -> Number:
        if isinstance(value, str):
            try:
                return int(value)
            except ValueError:
                return Decimal(value)
        else:
            return value


class NumberValidator(BoundedNumberValidator):
    def __init__(
        self,
        min_value: Optional[Number] = None,
        max_value: Optional[Number] = None,
        precision: Optional[int] = None,
    ) -> None:
        super().__init__(min_value=min_value, max_value=max_value)
        self.precision = precision

    def __call__(self, field: Field, value: Any) -> Any:
        try:
            value = self.to_number(value)
            if self.precision is not None:
                value = round(value, self.precision)
        except NumericErrors:
            raise ValidationError(field.name, "%s not valid number" % value)
        return super().__call__(field, value)

    def dump(self, value: Any) -> Any:
        value = self.to_number(value)
        if self.precision is not None:
            return round(value, self.precision)
        return value


class IntegerValidator(BoundedNumberValidator):
    def __call__(self, field: Field, value: Any) -> Any:
        try:
            value = self.to_number(value)
            if not isinstance(value, int):
                raise ValueError
        except NumericErrors:
            raise ValidationError(field.name, "%s not valid integer" % value)
        return super().__call__(field, value)


class DecimalValidator(NumberValidator):
    def __call__(self, field: Field, value: Any) -> Any:
        try:
            value = self.to_number(value)
            if not isinstance(value, Decimal):
                value = Decimal(str(value))
        except NumericErrors:
            raise ValidationError(field.name, "%s not valid Decimal" % value)
        return super().__call__(field, value)


class BoolValidator(Validator):
    def __call__(self, field: Field, value: Any) -> bool:
        return str2bool(value)

    def dump(self, value: Any) -> bool:
        return str2bool(value)


class JSONValidator(Validator):
    def __call__(self, field: Field, value: Any) -> Any:
        try:
            return self.dump(value)
        except (json.JSONDecodeError, TypeError):
            raise ValidationError(field.name, "%s not valid" % value)

    def dump(self, value: Any) -> Any:
        if isinstance(value, str):
            try:
                value = json.loads(value)
            except json.JSONDecodeError:
                pass
        return json.loads(json.dumps(value))



import os
from dataclasses import is_dataclass
from inspect import isclass
from typing import (
    Any,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    cast,
)

from .exc import InvalidTypeException
from .types import Record


def get_origin(value: Any) -> Any:
    return getattr(value, "__origin__", None)


LOCAL = "local"
DEV = "dev"
PRODUCTION = "production"
NO_DEBUG = {"0", "false", "no"}
Null = object()
#
# this should be Union[type, "TypingInfo"] but recursive types are not supported in mypy
ElementType = Any


def get_args(value, defaults):
    return getattr(value, "__args__", None) or defaults


KT, VT = get_args(Dict, (TypeVar("KT"), TypeVar("VT")))
(T,) = get_args(List, (TypeVar("T"),))


class TypingInfo(NamedTuple):
    """Information about a type annotation"""

    element: ElementType
    container: Optional[type] = None

    @property
    def is_dataclass(self) -> bool:
        """True if :attr:`.element` is a dataclass"""
        return not self.container and is_dataclass(self.element)

    @property
    def is_union(self) -> bool:
        """True if :attr:`.element` is a union of typing info"""
        return isinstance(self.element, tuple)

    @property
    def is_complex(self) -> bool:
        """True if :attr:`.element` is either a dataclass or a union"""
        return self.container is not None or self.is_union

    @property
    def is_none(self) -> bool:
        """True if :attr:`.element` is either a dataclass or a union"""
        return self.element is type(None)  # noqa: E721

    @classmethod
    def get(cls, value: Any) -> Optional["TypingInfo"]:
        """Create a :class:`.TypingInfo` from a typing annotation or
        another typing info

        :param value: typing annotation
        """
        if value is None or isinstance(value, cls):
            return value
        origin = get_origin(value)
        if not origin:
            if value is Any or isclass(value):
                return cls(value)
            else:
                raise InvalidTypeException(
                    f"a class or typing annotation is required, got {value}"
                )
        elif origin is list:
            (val,) = get_args(value, (T,))
            if val is T:
                val = Any
            elem_info = cast(TypingInfo, cls.get(val))
            elem = elem_info if elem_info.is_complex else elem_info.element
            return cls(elem, list)
        elif origin is dict:
            key, val = get_args(value, (KT, VT))
            if key is KT:
                key = str
            if val is VT:
                val = Any
            if key is not str:
                raise InvalidTypeException(
                    f"Dict key annotation must be a string, got {key}"
                )

            elem_info = cast(TypingInfo, cls.get(val))
            elem = elem_info if elem_info.is_complex else elem_info.element
            return cls(elem, dict)
        elif origin is Union:
            elem = tuple(cls.get(val) for val in value.__args__)
            return cls(elem)
        else:
            raise InvalidTypeException(
                f"Types or List and Dict typing is required, got {value}"
            )


def get_env() -> str:
    return os.environ.get("PYTHON_ENV") or PRODUCTION


def get_debug_flag() -> bool:
    val = os.environ.get("DEBUG")
    if not val:
        return get_env() == LOCAL
    return val.lower() not in NO_DEBUG


def compact(**kwargs) -> Dict:
    return {k: v for k, v in kwargs.items() if v}


def compact_dict(kwargs: Dict) -> Dict:
    return {k: v for k, v in kwargs.items() if v is not None}


def replace_key(kwargs: Dict, from_key: Hashable, to_key: Hashable) -> Dict:
    value = kwargs.pop(from_key, Null)
    if value is not Null:
        kwargs[to_key] = value
    return kwargs


def iter_items(data: Iterable) -> Iterator:
    if isinstance(data, Record):
        data = data._asdict()
    if isinstance(data, Mapping):
        return data.items()
    return iter(data)


def is_subclass(value: Any, Type: type) -> bool:
    origin = getattr(value, "__origin__", None) or value
    return isclass(origin) and issubclass(origin, Type)


def as_list(errors: Iterable) -> List[Dict[str, Any]]:
    return [
        {"field": field, "message": message} for field, message in iter_items(errors)
    ]


def error_dict(errors: List) -> Dict:
    return dict(((d["field"], d["message"]) for d in errors))


TRUE_VALUES = frozenset(("yes", "true", "t", "1"))


def str2bool(v: Union[str, bool, int]):
    return str(v).lower() in TRUE_VALUES


def docjoin(iterable: Iterable) -> str:
    return ", ".join(f"``{v}``" for v in iterable)



import asyncio
from functools import cached_property
from typing import Any, Callable, Dict, Set

from .channels import CannotSubscribe, Channels
from .errors import CannotPublish

WsHandlerType = Callable[[str, Any], None]


class Websocket:
    """A websocket connection"""

    socket_id: str = ""
    """websocket ID"""

    def __str__(self) -> str:
        return self.socket_id


class SocketsManager:
    """A base class for websocket managers"""

    @cached_property
    def sockets(self) -> Set[Websocket]:
        """Set of connected :class:`.Websocket`"""
        return set()

    @cached_property
    def channels(self) -> Channels:
        """Pub/sub :class:`.Channels` currently active on the running pod"""
        return Channels(self)

    def add(self, ws: Websocket) -> None:
        """Add a new websocket to the connected set"""
        self.sockets.add(ws)

    def remove(self, ws: Websocket) -> None:
        """Remove a websocket from the connected set"""
        self.sockets.discard(ws)

    def server_info(self) -> Dict:
        """Server information"""
        return dict(connections=len(self.sockets), channels=self.channels.info())

    async def close_sockets(self) -> None:
        """Close and remove all websockets from the connected set"""
        await asyncio.gather(*[view.response.close() for view in self.sockets])
        self.sockets.clear()
        self.channels.clear()

    async def publish(
        self, channel: str, event: str, body: Dict
    ) -> None:  # pragma: no cover
        """Publish an event to a channel

        :property channel: the channel to publish to
        :property event: the event in the channel
        :property body: the body of the event to broadcast in the channel

        This method should raise :class:`.CannotPublish` if not possible to publish
        """
        raise CannotPublish

    async def subscribe(self, channel: str) -> None:  # pragma: no cover
        """Subscribe to a channel

        This method should raise :class:`.CannotSubscribe` if not possible to publish
        """
        raise CannotSubscribe

    async def subscribe_to_event(self, channel: str, event: str) -> None:
        """Callback when a subscription to an event is done

        :property channel: the channel to publish to
        :property event: the event in the channel

        You can use this callback to perform any backend subscriptions to
        third-party streaming services if required.

        By default it does nothing.
        """

    async def unsubscribe(self, channel: str) -> None:
        """Unsubscribe from a channel"""



from aiohttp.web import HTTPException

from .json import dumps


class OpenApiError(RuntimeError):
    pass


class ImproperlyConfigured(OpenApiError):
    pass


class InvalidSpecException(OpenApiError):
    pass


class InvalidTypeException(TypeError):
    pass


class JsonHttpException(HTTPException):
    def __init__(self, status=None, **kw):
        self.status_code = status or 500
        kw["content_type"] = "application/json"
        super().__init__(**kw)
        reason = self.reason
        if isinstance(reason, str):
            reason = {"message": reason}
        self.text = dumps(reason)



import re
from typing import List, Optional, Tuple, cast

import sqlalchemy as sa
from aiohttp import web
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import Select

from openapi.data.validate import ValidationErrors

from ..pagination import PaginatedData, Pagination, Search, create_dataclass
from ..spec.path import ApiPath
from ..types import Connection, DataType, Record, Records, SchemaTypeOrStr, StrDict
from .dbmodel import CrudDB

unique_regex = re.compile(r"Key \((?P<column>(\w+,? ?)+)\)=\((?P<value>.+)\)")


class SqlApiPath(ApiPath):
    """An :class:`.ApiPath` backed by an SQL model.

    This class provides utility methods for all CRUD operations.
    """

    table: str = ""
    """sql table name"""

    @property
    def db(self) -> CrudDB:
        """Database connection pool"""
        return self.request.app["db"]

    @property
    def db_table(self) -> sa.Table:
        """Default database table for this route.

        Obtained from the :attr:`table` attribute.
        """
        return self.db.metadata.tables[self.table]

    def get_pag_search_filters(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
    ) -> Tuple[Pagination, Search, dict]:
        if filters is None:
            filters = self.get_filters(query=query, query_schema=query_schema)
        schema = self.get_schema(query_schema)
        pagination = create_dataclass(schema, filters, Pagination)
        search = create_dataclass(schema, filters, Search)
        return pagination, search, filters

    async def get_list(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ) -> PaginatedData:
        """Get a list of models

        :param filters: dictionary of filters, if not provided it will be created from
            the query_schema
        :param query: additional query parameters, only used when filters
            is not provided
        :param table: sqlalchemy table, if not provided the default :attr:`.db_table` is
            used instead
        :param conn: optional db connection
        """
        table = table if table is not None else self.db_table
        pagination, search, filters = self.get_pag_search_filters(
            filters=filters, query=query, query_schema=query_schema
        )
        sql_query = cast(
            Select,
            self.db.get_query(
                table,
                table.select(),
                params=filters,
                consumer=self,
            ),
        )
        sql_query = cast(Select, self.db.search_query(table, sql_query, search))
        try:
            values, total = await self.db.db_paginate(
                table, sql_query, pagination, conn=conn
            )
        except ValidationErrors as e:
            self.raise_validation_error(errors=e.errors)
        data = cast(List[StrDict], self.dump(dump_schema, values.all()))
        return pagination.paginated(self.full_url(), data, total)

    async def create_one(
        self,
        *,
        data: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        body_schema: SchemaTypeOrStr = "body_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ) -> StrDict:
        """Create a new database model

        :param data: input data, if not given it loads it via :meth:`.json_data`
        :param table: sqlalchemy table, if not given it uses the
            default :attr:`db_table`
        :param conn: optional db connection
        """
        if data is None:
            data = self.insert_data(await self.json_data(), body_schema=body_schema)
        table = table if table is not None else self.db_table
        sql = self.db.insert_query(table, data)

        async with self.db.ensure_connection(conn) as conn:
            try:
                result = await conn.execute(sql)
            except IntegrityError as exc:
                self.handle_unique_violation(exc)
                raise

        return cast(StrDict, self.dump(dump_schema, result.one()))

    async def create_list(
        self,
        *,
        data: Optional[DataType] = None,
        table: Optional[sa.Table] = None,
        body_schema: SchemaTypeOrStr = "body_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ):
        """Create multiple models"""
        table = table if table is not None else self.db_table
        if data is None:
            data = await self.json_data()
        if not isinstance(data, list):
            raise web.HTTPBadRequest(
                **self.api_response_data({"message": "Invalid JSON payload"})
            )
        schema = self.get_schema(body_schema)
        assert schema.container is list
        data = [self.insert_data(d, body_schema=schema.element) for d in data]
        values = await self.db.db_insert(table, data, conn=conn)
        return self.dump(dump_schema, values.all())

    async def get_one(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ):
        """Get a single model

        :param filters: dictionary of filters, if not provided it will be created from
            the query_schema
        :param query: additional query parameters, only used when filters
            is not provided
        :param table: sqlalchemy table, if not provided the default :attr:`.db_table` is
            used instead
        :param conn: optional db connection
        """
        table = table if table is not None else self.db_table
        if filters is None:
            filters = self.get_filters(query=query, query_schema=query_schema)
        values = await self.db.db_select(table, filters, conn=conn, consumer=self)
        row = values.first()
        if row is None:
            raise web.HTTPNotFound()
        return self.dump(dump_schema, row)

    async def update_one(
        self,
        *,
        data: Optional[StrDict] = None,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        body_schema: SchemaTypeOrStr = "body_schema",
        query_schema: SchemaTypeOrStr = "query_schema",
        dump_schema: SchemaTypeOrStr = "response_schema",
        conn: Optional[Connection] = None,
    ):
        """Update a single model"""
        table = table if table is not None else self.db_table
        if data is None:
            data = self.cleaned("body_schema", await self.json_data(), strict=False)
        if not filters:
            filters = self.get_filters(query=query, query_schema=query_schema)

        if data:
            try:
                values = await self.db.db_update(
                    table, filters, data, conn=conn, consumer=self
                )
            except IntegrityError as exc:
                self.handle_unique_violation(exc)
                raise
        else:
            values = await self.db.db_select(table, filters, conn=conn, consumer=self)
        row = values.first()
        if row is None:
            raise web.HTTPNotFound()
        return self.dump(dump_schema, row)

    async def delete_one(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
        conn: Optional[Connection] = None,
    ) -> Record:
        """Delete a single model"""
        table = table if table is not None else self.db_table
        if filters is None:
            filters = self.get_filters(query=query, query_schema=query_schema)
        values = await self.db.db_delete(table, filters, conn=conn, consumer=self)
        row = values.first()
        if row is None:
            raise web.HTTPNotFound()
        return row

    async def delete_list(
        self,
        *,
        filters: Optional[StrDict] = None,
        query: Optional[StrDict] = None,
        table: Optional[sa.Table] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
        conn: Optional[Connection] = None,
    ) -> Records:
        """delete multiple models"""
        table = table if table is not None else self.db_table
        # TODO: allow pagination and search
        _, _, filters = self.get_pag_search_filters(
            filters=filters, query=query, query_schema=query_schema
        )
        return await self.db.db_delete(table, filters, conn=conn, consumer=self)

    def handle_unique_violation(self, exception: IntegrityError):
        match = re.search(unique_regex, str(exception))
        if match:
            column = match.group("column")
            message = f"{column} already exists"
            self.raise_validation_error(message=message)



import os
from dataclasses import dataclass, fields
from typing import (
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from aiohttp import web
from yarl import URL

from openapi.json import dumps

MAX_PAGINATION_LIMIT: int = int(os.environ.get("MAX_PAGINATION_LIMIT") or 100)
DEF_PAGINATION_LIMIT: int = int(os.environ.get("DEF_PAGINATION_LIMIT") or 50)


class PaginationVisitor:
    """Visitor for pagination"""

    def apply_offset_pagination(
        self, limit: int, offset: int, order_by: Union[str, List[str]]
    ):
        raise NotImplementedError

    def apply_cursor_pagination(
        self,
        cursor: Sequence[Tuple[str, str]],
        limit: int,
        order_by: Sequence[str],
        previous: bool,
    ):
        raise NotImplementedError


T = TypeVar("T")


def from_filters_and_dataclass(data_class: Type[T], data: dict) -> T:
    params = {}
    for field in fields(data_class):
        if field.name in data:
            params[field.name] = data.pop(field.name)
    return data_class(**params)


def fields_no_sign(fields: Sequence[str]) -> Tuple[str, ...]:
    return tuple(field[1:] if field.startswith("-") else field for field in fields)


def fields_flip_sign(fields: Sequence[str]) -> Tuple[str, ...]:
    return tuple(flip_field_sign(field) for field in fields)


def flip_field_sign(field: str) -> str:
    return field[1:] if field.startswith("-") else f"-{field}"


@dataclass
class Pagination:
    @classmethod
    def create_pagination(cls, data: dict) -> "Pagination":
        return cls()

    def apply(self, visitor: PaginationVisitor) -> None:
        """Apply pagination to the visitor"""
        pass

    def paginated(
        self, url: URL, data: list, total: Optional[int] = None
    ) -> "PaginatedData":
        """Return paginated data"""
        return PaginatedData(url=url, data=data, pagination=self, total=total)

    def links(
        self, url: URL, data: list, total: Optional[int] = None
    ) -> Dict[str, str]:
        """Return links for paginated data"""
        return {}

    def get_data(self, data: list) -> list:
        return data


class PaginatedData(NamedTuple):
    url: URL
    data: list
    pagination: Pagination
    total: Optional[int] = None

    def json_response(self, headers: Optional[Dict[str, str]] = None, **kwargs):
        headers = headers or {}
        links = self.header_links()
        if links:
            headers["Link"] = links
        if self.total is not None:
            headers["X-Total-Count"] = str(self.total)
        kwargs.setdefault("dumps", dumps)
        return web.json_response(
            self.pagination.get_data(self.data), headers=headers, **kwargs
        )

    def header_links(self) -> str:
        links = self.pagination.links(self.url, self.data, self.total)
        return ", ".join(f'<{value}>; rel="{name}"' for name, value in links.items())



from typing import Any, Dict, Optional

from aiohttp import web
from multidict import MultiDict
from yarl import URL

from openapi.json import dumps, loads

from ..data.validate import ValidationErrors
from ..data.view import BAD_DATA_MESSAGE, DataView, ErrorType
from ..types import DataType, QueryType, SchemaTypeOrStr
from ..utils import compact
from . import hdrs


class ApiPath(web.View, DataView):
    """A :class:`.DataView` class for OpenAPI path"""

    path_schema: Optional[type] = None
    """Optional dataclass for validating path variables"""
    private: bool = False

    # UTILITIES

    def insert_data(
        self,
        data: DataType,
        *,
        multiple: bool = False,
        strict: bool = True,
        body_schema: SchemaTypeOrStr = "body_schema",
    ) -> Dict[str, Any]:
        """Validate data for insertion

        if a :attr:`.path_schema` is given, it validate the request `match_info`
        against it and add it to the validated data.

        :param data: object to be validated against the body_schema, usually obtained
            from the request body (JSON)
        :param multiple: multiple values for a given key are acceptable
        :param strict: all required attributes in schema must be available
        :param body_schema: the schema to validate against
        """
        data = self.cleaned(body_schema, data, multiple=multiple, strict=strict)
        if self.path_schema:
            path = self.cleaned("path_schema", self.request.match_info)
            data.update(path)
        return data

    def get_filters(
        self,
        *,
        query: Optional[QueryType] = None,
        query_schema: SchemaTypeOrStr = "query_schema",
    ) -> Dict[str, Any]:
        """Collect a dictionary of filters from the request query string.
        If :attr:`path_schema` is defined, it collects filter data
        from path variables as well.

        :param query: optional query dictionary (will be overwritten
            by the request.query)
        :param query_schema: a dataclass or an the name of an attribute in Operation
            for collecting query filters
        """
        combined = MultiDict(query or ())
        combined.update(self.request.query)
        try:
            params = self.cleaned(query_schema, combined, multiple=True)
        except web.HTTPNotImplemented:
            params = {}
        if self.path_schema:
            path = self.cleaned("path_schema", self.request.match_info)
            params.update(path)
        return params

    async def json_data(self) -> DataType:
        """Load JSON data from the request.

        :raise HTTPBadRequest: when body data is not valid JSON
        """
        try:
            return await self.request.json(loads=loads)
        except Exception:
            self.raise_bad_data()

    def validation_error(
        self, message: str = "", errors: Optional[ErrorType] = None
    ) -> Exception:
        """Create an :class:`aiohttp.web.HTTPUnprocessableEntity`"""
        raw = self.as_errors(message, errors)
        data = self.dump(ValidationErrors, raw)
        return web.HTTPUnprocessableEntity(**self.api_response_data(data))

    def raise_bad_data(
        self, exc: Optional[Exception] = None, message: str = ""
    ) -> None:
        raw = compact(message=message or BAD_DATA_MESSAGE)
        data = self.dump(ValidationErrors, raw)
        raise web.HTTPBadRequest(**self.api_response_data(data))

    def full_url(self) -> URL:
        return full_url(self.request)

    @classmethod
    def api_response_data(cls, data: DataType) -> Dict[str, Any]:
        return dict(text=dumps(data), content_type="application/json")

    @classmethod
    def json_response(cls, data, **kwargs):
        kwargs.setdefault("dumps", dumps)
        return web.json_response(data, **kwargs)


def full_url(request) -> URL:
    headers = request.headers
    proto = headers.get(hdrs.X_FORWARDED_PROTO)
    host = headers.get(hdrs.X_FORWARDED_HOST)
    port = headers.get(hdrs.X_FORWARDED_PORT)
    if proto and host:
        url = URL.build(scheme=proto, host=host)
        if port:
            port = int(port)
            if url.port != port:
                url = url.with_port(port)
        return url.join(request.rel_url)
    else:
        return request.url



from .create import create_dataclass
from .cursor import cursorPagination
from .offset import offsetPagination
from .pagination import PaginatedData, Pagination, PaginationVisitor, fields_flip_sign
from .search import Search, SearchVisitor, searchable

__all__ = [
    "Pagination",
    "PaginatedData",
    "PaginationVisitor",
    "cursorPagination",
    "offsetPagination",
    "create_dataclass",
    "fields_flip_sign",
    "Search",
    "searchable",
    "SearchVisitor",
]



from dataclasses import MISSING, Field, fields
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union, cast

from multidict import MultiDict

from openapi import json
from openapi.types import Record

from ..utils import TypingInfo
from .fields import (
    ITEMS,
    POST_PROCESS,
    REQUIRED,
    VALIDATOR,
    ValidationError,
    as_field,
    field_ops,
)

NOT_VALID_TYPE = "not valid type"
OBJECT_EXPECTED = "expected an object"
VALIDATION_SIMPLE_MAP = {float: (float, int)}

ErrorType = Union[Dict, str, None]


class ValidatedData(NamedTuple):
    data: Any = None
    errors: ErrorType = None


class ValidationErrors(ValueError):
    def __init__(self, errors: Union[Dict, str]) -> None:
        self.errors = errors

    def __repr__(self) -> str:
        return (
            self.errors
            if isinstance(self.errors, str)
            else json.dumps(self.errors, indent=4)
        )

    __str__ = __repr__


def validated_schema(
    schema: Any,
    data: Any,
    *,
    multiple: bool = False,
    strict: bool = True,
) -> Any:
    """Validate data with a given schema and return a valid representation of the data
    as a schema instance

    :param schema: a valid  :ref:`aio-openapi-schema` or a :class:`.TypingInfo` object
    :param data: a data object to validate against the schema
    :param strict: if `True` validation is strict, i.e. missing required parameters
        will cause validation to fails
    """
    return validate(
        schema,
        data,
        strict=strict,
        multiple=multiple,
        raise_on_errors=True,
        as_schema=True,
    )


def validate(
    schema: Any,
    data: Any,
    *,
    strict: bool = True,
    multiple: bool = False,
    raise_on_errors: bool = False,
    items: Optional[Field] = None,
    as_schema: bool = False,
) -> Any:
    """Validate data with a given schema

    :param schema: a typing annotation or a :class:`.TypingInfo` object
    :param data: a data object to validate against the schema
    :param strict: if `True` validation is strict, i.e. missing required parameters
        will cause validation to fails
    :param multiple: allow parameters to have multiple values
    :param raise_on_errors: when `True` failure of validation will result in a
        `ValidationErrors` error, otherwise a :class:`.ValidatedData` object
        is returned.
    :param items: an optional Field for items in a composite type (`List` or `Dict`)
    :param as_schema: return the schema object rather than simple data type
        (dataclass rather than dict for example)
    """
    type_info = cast(TypingInfo, TypingInfo.get(schema))
    try:
        if type_info.container is list:
            vdata = validate_list(
                type_info.element,
                data,
                strict=strict,
                multiple=multiple,
                items=items,
                as_schema=as_schema,
            )
        elif type_info.container is dict:
            vdata = validate_dict(
                type_info.element,
                data,
                strict=strict,
                multiple=multiple,
                items=items,
                as_schema=as_schema,
            )
        elif type_info.is_dataclass:
            vdata = validate_dataclass(
                type_info.element,
                data,
                strict=strict,
                multiple=multiple,
                as_schema=as_schema,
            )
        elif type_info.is_union:
            vdata = validate_union(type_info.element, data, as_schema=as_schema)
        elif type_info.element is Any:
            vdata = data
        else:
            vdata = validate_simple(type_info.element, data)
    except ValidationErrors as e:
        if not raise_on_errors:
            return ValidatedData(errors=e.errors)
        raise
    else:
        return vdata if raise_on_errors else ValidatedData(data=vdata, errors={})


def validate_simple(schema: type, data: Any) -> Any:
    if isinstance(data, VALIDATION_SIMPLE_MAP.get(schema, schema)):
        return data
    raise ValidationErrors(NOT_VALID_TYPE)


def validate_union(
    schema: Tuple[TypingInfo, ...],
    data: Any,
    as_schema: bool = False,
    **kw,
) -> Any:
    for type_info in schema:
        try:
            return validate(type_info, data, raise_on_errors=True, as_schema=as_schema)
        except ValidationErrors:
            continue
    raise ValidationErrors(NOT_VALID_TYPE)


def validate_list(
    schema: type,
    data: list,
    *,
    strict: bool = True,
    multiple: bool = False,
    as_schema: bool = False,
    items: Optional[Field] = None,
) -> ValidatedData:
    validated = []
    if isinstance(data, (list, tuple)):
        items = as_field(schema, field=items)
        for d in data:
            v = collect_value(
                items,
                d,
                strict=strict,
                multiple=multiple,
                as_schema=as_schema,
            )
            validated.append(v)
        return validated
    else:
        raise ValidationErrors("expected a sequence")


def validate_dict(
    schema: type,
    data: Dict[str, Any],
    *,
    strict: bool = True,
    multiple: bool = False,
    as_schema: bool = False,
    items: Optional[Field] = None,
) -> ValidatedData:
    if isinstance(data, dict):
        validated = ValidatedData(data={}, errors={})
        items = as_field(schema, field=items)
        for name, d in data.items():
            try:
                validated.data[name] = collect_value(
                    items,
                    d,
                    strict=strict,
                    multiple=multiple,
                    as_schema=as_schema,
                )
            except ValidationErrors as exc:
                validated.errors[name] = exc.errors
        if validated.errors:
            raise ValidationErrors(validated.errors)
        return validated.data
    else:
        raise ValidationErrors(OBJECT_EXPECTED)


def validate_dataclass(
    schema: type,
    data: Union[Dict[str, Any], MultiDict, Record],
    *,
    strict: bool = True,
    multiple: bool = False,
    as_schema: bool = False,
    **kw,
) -> ValidatedData:
    errors: Dict = {}
    cleaned: Dict = {}
    try:
        data = MultiDict(dict(data) if isinstance(data, Record) else data)
    except TypeError:
        raise ValidationErrors(OBJECT_EXPECTED)
    for field in fields(schema):
        try:
            required = field.metadata.get(REQUIRED, True)
            default = get_default(field)
            if strict and default is not None and data.get(field.name) is None:
                data[field.name] = default

            if field.name not in data and required and strict:
                raise ValidationError(field.name, "required")

            for name in field_ops(field):
                if name not in data:
                    continue

                if multiple:
                    values = data.getall(name)
                    if len(values) > 1:
                        collected = []
                        for v in values:
                            v = collect_value(field, v)
                            if v is not None:
                                collected.append(v)
                        value = collected if collected else None
                    else:
                        value = collect_value(field, values[0], as_schema=as_schema)
                else:
                    value = collect_value(field, data[name], as_schema=as_schema)

                cleaned[name] = value

        except ValidationError as exc:
            errors[exc.field] = exc.message
        except ValidationErrors as exc:
            errors[name] = exc.errors

    if not errors:
        validate = getattr(schema, "validate", None)
        if validate:
            validate(cleaned, errors)

    if errors:
        raise ValidationErrors(errors)
    return schema(**cleaned) if as_schema else cleaned


def collect_value(field: Field, value: Any, **kw) -> Any:
    if is_null(value):
        return None

    validator = field.metadata.get(VALIDATOR)
    if validator:
        value = validator(field, value)

    kw.update(raise_on_errors=True, items=field.metadata.get(ITEMS))
    value = validate(field.type, value, **kw)

    post_process = field.metadata.get(POST_PROCESS)
    return post_process(value) if post_process else value


def is_null(value: Any) -> bool:
    return value is None or value == "NULL"


def get_default(field: Field) -> Any:
    if field.default_factory is not MISSING:
        value = field.default_factory()
    else:
        value = field.default
    return value if value is not MISSING else None



import re
from typing import Dict, Optional

import yaml

from ..exc import InvalidSpecException


# from django.contrib.admindocs.utils
def trim_docstring(docstring: str) -> str:
    """Uniformly trims leading/trailing whitespace from docstrings.
    Based on
    http://www.python.org/peps/pep-0257.html#handling-docstring-indentation
    """
    if not docstring or not docstring.strip():
        return ""
    # Convert tabs to spaces and split into lines
    lines = docstring.expandtabs().splitlines()
    indent = min(len(line) - len(line.lstrip()) for line in lines if line.lstrip())
    trimmed = [lines[0].lstrip()] + [line[indent:].rstrip() for line in lines[1:]]
    return "\n".join(trimmed).strip()


# from rest_framework.utils.formatting
def dedent(content: str) -> str:
    """
    Remove leading indent from a block of text.
    Used when generating descriptions from docstrings.
    Note that python's `textwrap.dedent` doesn't quite cut it,
    as it fails to dedent multiline docstrings that include
    unindented text on the initial line.
    """
    whitespace_counts = [
        len(line) - len(line.lstrip(" "))
        for line in content.splitlines()[1:]
        if line.lstrip()
    ]

    # unindent the content if needed
    if whitespace_counts:
        whitespace_pattern = "^" + (" " * min(whitespace_counts))
        content = re.sub(re.compile(whitespace_pattern, re.MULTILINE), "", content)

    return content.strip()


def load_yaml_from_docstring(docstring: str) -> Optional[Dict]:
    """Loads YAML from docstring."""
    split_lines = trim_docstring(docstring).split("\n")

    # Cut YAML from rest of docstring
    for index, line in enumerate(split_lines):
        line = line.strip()
        if line.startswith("---"):
            cut_from = index
            break
    else:
        return None

    yaml_string = "\n".join(split_lines[cut_from:])
    yaml_string = dedent(yaml_string)
    try:
        return yaml.load(yaml_string, Loader=yaml.FullLoader)
    except Exception as e:
        raise InvalidSpecException("Invalid yaml %s" % e) from None



import click
from sqlalchemy_utils import create_database, database_exists, drop_database

from openapi.cli import open_api_cli

from .dbmodel import CrudDB
from .migrations import Migration


def migration(ctx: click.Context) -> Migration:
    return Migration(open_api_cli(ctx).web())


def get_db(ctx: click.Context) -> CrudDB:
    return open_api_cli(ctx).web()["db"]


@click.group()
def db():
    """Perform database migrations and utilities"""
    pass


@db.command()
@click.pass_context
def init(ctx):
    """Creates a new migration repository."""
    migration(ctx).init()


@db.command()
@click.option("-m", "--message", help="Revision message", required=True)
@click.option(
    "--branch-label", help="Specify a branch label to apply to the new revision"
)
@click.pass_context
def migrate(ctx, message: str, branch_label: str):
    """Autogenerate a new revision file

    alias for 'revision --autogenerate'
    """
    return migration(ctx).revision(
        message, autogenerate=True, branch_label=branch_label
    )


@db.command()
@click.option("-m", "--message", help="Revision message", required=True)
@click.option(
    "--branch-label", help="Specify a branch label to apply to the new revision"
)
@click.option(
    "--autogenerate",
    default=False,
    is_flag=True,
    help=(
        "Populate revision script with candidate migration "
        "operations, based on comparison of database to model"
    ),
)
@click.pass_context
def revision(ctx, message: str, branch_label: str, autogenerate: bool):
    """Autogenerate a new revision file"""
    return migration(ctx).revision(
        message, autogenerate=autogenerate, branch_label=branch_label
    )


@db.command()
@click.option("--revision", default="heads")
@click.option(
    "--drop-tables",
    default=False,
    is_flag=True,
    help="Drop tables before applying migrations",
)
@click.pass_context
def upgrade(ctx, revision: str, drop_tables: bool):
    """Upgrade to a later version"""
    if drop_tables:
        _drop_tables(ctx)
    migration(ctx).upgrade(revision)
    click.echo(f"upgraded successfully to {revision}")


@db.command()
@click.option("--revision", help="Revision id", required=True)
@click.pass_context
def downgrade(ctx, revision: str):
    """Downgrade to a previous version"""
    migration(ctx).downgrade(revision)
    click.echo(f"downgraded successfully to {revision}")


@db.command()
@click.option("--revision", default="heads")
@click.pass_context
def show(ctx, revision: str):
    """Show revision ID and creation date"""
    click.echo(migration(ctx).show(revision))


@db.command()
@click.pass_context
def history(ctx):
    """List changeset scripts in chronological order"""
    click.echo(migration(ctx).history())


@db.command()
@click.option("--verbose/--quiet", default=False)
@click.pass_context
def current(ctx, verbose: bool):
    """Show revision ID and creation date"""
    click.echo(migration(ctx).current(verbose))


@db.command()
@click.argument("dbname", nargs=1)
@click.option(
    "--force", default=False, is_flag=True, help="Force removal of an existing database"
)
@click.pass_context
def create(ctx, dbname: str, force: str):
    """Creates a new database"""
    engine = get_db(ctx).sync_engine
    url = engine.url.set(database=dbname)
    store = str(url)
    if database_exists(store):
        if force:
            drop_database(store)
        else:
            return click.echo(f"database {dbname} already available")
    create_database(store)
    click.echo(f"database {dbname} created")


@db.command()
@click.option(
    "--db",
    default=False,
    is_flag=True,
    help="List tables in database rather than in sqlalchemy metadata",
)
@click.pass_context
def tables(ctx, db):
    """List all tables managed by the app"""
    d = get_db(ctx)
    if db:
        tables = d.sync_engine.table_names()
    else:
        tables = d.metadata.tables
    for name in sorted(tables):
        click.echo(name)


@db.command()
@click.pass_context
def drop(ctx):
    """Drop all tables in database"""
    _drop_tables(ctx)


def _drop_tables(ctx):
    get_db(ctx).drop_all_schemas()
    click.echo("tables dropped")



import asyncio
import os
import sys
from functools import lru_cache
from typing import Callable, Iterable, List, Optional

import click
from aiohttp import web
from aiohttp.web import Application

from .logger import logger, setup_logging
from .spec import OpenApiSpec
from .utils import get_debug_flag

HOST = os.environ.get("MICRO_SERVICE_HOST", "0.0.0.0")
PORT = os.environ.get("MICRO_SERVICE_PORT", 8080)


class OpenApiClient(click.Group):
    def __init__(
        self,
        spec: Optional[OpenApiSpec] = None,
        setup_app: Optional[Callable[[Application], None]] = None,
        base_path: str = "",
        commands: Optional[List] = None,
        index: int = -1,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        **extra,
    ) -> None:
        params = list(extra.pop("params", None) or ())
        self.spec = spec
        self.debug = get_debug_flag()
        self.setup_app = setup_app
        self.base_path: str = base_path or ""
        self.index = index
        self.loop = loop
        params.extend(
            (
                click.Option(
                    ["--version"],
                    help="Show the server version",
                    expose_value=False,
                    callback=self.get_server_version,
                    is_flag=True,
                    is_eager=True,
                ),
                click.Option(
                    ["-v", "--verbose"],
                    help="Increase logging verbosity",
                    is_flag=True,
                    is_eager=True,
                ),
                click.Option(
                    ["-q", "--quiet"],
                    help="Decrease logging verbosity",
                    is_flag=True,
                    is_eager=True,
                ),
            )
        )
        extra.setdefault("callback", setup_logging)
        super().__init__(params=params, **extra)
        self.add_command(serve)
        for command in commands or ():
            self.add_command(command)

    @lru_cache(None)
    def web(self, server: bool = False) -> Application:
        """Return the web application"""
        app = Application()
        app["cli"] = self
        app["cwd"] = os.getcwd()
        app["index"] = self.index
        app["server"] = server
        if self.spec:
            self.spec.setup_app(app)
        if self.setup_app:
            self.setup_app(app)
        return app

    def get_serve_app(self) -> Application:
        """Create the application which runs the server"""
        app = self.web(server=True)
        if self.base_path:
            base = Application()
            base.add_subapp(self.base_path, app)
            base["cli"] = self
            app = base
        return app

    def get_command(self, ctx: click.Context, name: str) -> Optional[click.Command]:
        ctx.obj = dict(cli=self)
        return super().get_command(ctx, name)

    def list_commands(self, ctx: click.Context) -> Iterable[str]:
        ctx.obj = dict(cli=self)
        return super().list_commands(ctx)

    def get_server_version(self, ctx, param, value) -> None:
        if not value or ctx.resilient_parsing:
            return
        spec = self.spec
        message = "%(title)s %(version)s\nPython %(python_version)s"
        click.echo(
            message
            % {
                "title": spec.info.title if spec else self.name or "Open API",
                "version": spec.info.version if spec else "",
                "python_version": sys.version,
            },
            color=ctx.color,
        )
        ctx.exit()


def open_api_cli(ctx: click.Context) -> OpenApiClient:
    return ctx.obj["cli"]


@click.command("serve", short_help="Start aiohttp server.")
@click.option(
    "--host", "-h", default=HOST, help=f"The interface to bind to (default to {HOST})"
)
@click.option(
    "--port", "-p", default=PORT, help=f"The port to bind to (default to {PORT}."
)
@click.option(
    "--index",
    default=0,
    type=int,
    help="Optional index for stateful set deployment",
)
@click.option(
    "--reload/--no-reload",
    default=None,
    help="Enable or disable the reloader. By default the reloader "
    "is active if debug is enabled.",
)
@click.pass_context
def serve(ctx, host, port, index, reload):
    """Run the aiohttp server."""
    cli = open_api_cli(ctx)
    cli.index = index
    app = cli.get_serve_app()
    access_log = logger if ctx.obj["log_level"] else None
    web.run_app(
        app,
        host=host,
        port=port,
        access_log=access_log,
        loop=cli.loop,
        print=access_log.info if access_log else None,
    )



from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable

from ..data.view import DataView, Operation
from ..utils import TypingInfo


@dataclass
class op:
    """Decorator for a :class:`.ApiPath` view which specifies an operation object
    in an OpenAPI Path. Parameters are dataclasses used for validation and
    OpenAPI auto documentation.
    """

    body_schema: Any = None
    query_schema: Any = None
    response_schema: Any = None
    response: int = 200
    # responses: List[Any] = []

    def __call__(self, method) -> Callable:
        method.op = Operation(
            body_schema=TypingInfo.get(self.body_schema),
            query_schema=TypingInfo.get(self.query_schema),
            response_schema=TypingInfo.get(self.response_schema),
            response=self.response,
        )

        @wraps(method)
        async def _(view: DataView) -> Any:
            view.operation = method.op
            return await method(view)

        return _



"""Alembic migrations handler
"""
import os
from io import StringIO

from alembic import command as alembic_cmd
from alembic.config import Config


def get_template_directory():
    return os.path.dirname(os.path.realpath(__file__))


class Migration:
    def __init__(self, app):
        self.app = app
        self.cfg = create_config(app)

    def init(self):
        dirname = self.cfg.get_main_option("script_location")
        alembic_cmd.init(self.cfg, dirname, template="openapi")
        return self.message()

    def show(self, revision):
        alembic_cmd.show(self.cfg, revision)
        return self.message()

    def history(self):
        alembic_cmd.history(self.cfg)
        return self.message()

    def revision(self, message, autogenerate=False, branch_label=None):
        alembic_cmd.revision(
            self.cfg,
            autogenerate=autogenerate,
            message=message,
            branch_label=branch_label,
        )
        return self.message()

    def upgrade(self, revision):
        alembic_cmd.upgrade(self.cfg, revision)
        return self.message()

    def downgrade(self, revision):
        alembic_cmd.downgrade(self.cfg, revision)
        return self.message()

    def current(self, verbose=False):
        alembic_cmd.current(self.cfg, verbose=verbose)
        return self.message()

    def message(self):
        msg = self.cfg.stdout.getvalue()
        self.cfg.stdout.seek(0)
        self.cfg.stdout.truncate()
        return msg


def create_config(app):
    """Programmatically create Alembic config"""
    cfg = Config(stdout=StringIO())
    cfg.get_template_directory = get_template_directory
    migrations = app.get("migrations_dir") or os.path.join(app["cwd"], "migrations")
    cfg.set_main_option("script_location", migrations)
    cfg.config_file_name = os.path.join(migrations, "alembic.ini")
    db = app["db"]
    cfg.set_section_option("default", "sqlalchemy.url", str(db.sync_engine.url))
    # put database in main options
    cfg.set_main_option("databases", "default")
    # create empty logging section to avoid raising errors in env.py
    cfg.set_section_option("logging", "path", "")
    cfg.metadata = dict(default=db.metadata)
    return cfg



import os
from contextlib import asynccontextmanager
from typing import Any, Optional

import sqlalchemy as sa
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

from openapi.types import Connection
from openapi.utils import str2bool

from ..exc import ImproperlyConfigured

DBPOOL_MAX_SIZE = int(os.environ.get("DBPOOL_MAX_SIZE") or "10")
DBPOOL_MAX_OVERFLOW = int(os.environ.get("DBPOOL_MAX_OVERFLOW") or "10")
DBECHO = str2bool(os.environ.get("DBECHO") or "no")


class Database:
    """A container for tables in a database and a manager of asynchronous
    connections to a postgresql database

    :param dsn: Data source name used for database connections
    :param metadata: :class:`sqlalchemy.schema.MetaData` containing tables
    """

    def __init__(self, dsn: str = "", metadata: sa.MetaData = None) -> None:
        self._dsn = dsn
        self._metadata = metadata or sa.MetaData()
        self._engine = None

    def __repr__(self) -> str:
        return self._dsn

    __str__ = __repr__

    @property
    def dsn(self) -> str:
        """Data source name used for database connections"""
        return self._dsn

    @property
    def metadata(self) -> sa.MetaData:
        """The :class:`sqlalchemy.schema.MetaData` containing tables"""
        return self._metadata

    @property
    def engine(self) -> AsyncEngine:
        """The :class:`sqlalchemy.ext.asyncio.AsyncEngine` creating connection
        and transactions"""
        if self._engine is None:
            if not self._dsn:
                raise ImproperlyConfigured("DSN not available")
            self._engine = create_async_engine(
                self._dsn,
                echo=DBECHO,
                pool_size=DBPOOL_MAX_SIZE,
                max_overflow=DBPOOL_MAX_OVERFLOW,
            )
        return self._engine

    @property
    def sync_engine(self) -> Engine:
        """The :class:`sqlalchemy.engine.Engine` for synchrouns operations"""
        return create_engine(self._dsn.replace("+asyncpg", ""))

    def __getattr__(self, name: str) -> Any:
        """Retrive a :class:`sqlalchemy.schema.Table` from metadata tables

        :param name: if this is a valid table name in the tables of :attr:`.metadata`
            it returns the table, otherwise it defaults to superclass method
        """
        if name in self._metadata.tables:
            return self._metadata.tables[name]
        return super().__getattribute__(name)

    @asynccontextmanager
    async def connection(self) -> Connection:
        """Context manager for obtaining an asynchronous connection"""
        async with self.engine.connect() as conn:
            yield conn

    @asynccontextmanager
    async def transaction(self) -> Connection:
        """Context manager for initializing an asynchronous database transaction"""
        async with self.engine.begin() as conn:
            yield conn

    @asynccontextmanager
    async def ensure_connection(self, conn: Optional[Connection] = None) -> Connection:
        """Context manager for ensuring we a connection has initialized
        a database transaction"""
        if conn:
            if not conn.in_transaction():
                async with conn.begin():
                    yield conn
            else:
                yield conn
        else:
            async with self.transaction() as conn:
                yield conn

    async def close(self) -> None:
        """Close the asynchronous db engine if opened"""
        if self._engine:
            engine, self._engine = self._engine, None
            await engine.dispose()

    # SQL Alchemy Sync Operations
    def create_all(self) -> None:
        """Create all tables defined in :attr:`metadata`"""
        self.metadata.create_all(self.sync_engine)

    def drop_all(self) -> None:
        """Drop all tables from :attr:`metadata` in database"""
        with self.sync_engine.begin() as conn:
            conn.execute(sa.text(f'truncate {", ".join(self.metadata.tables)}'))
            try:
                conn.execute(sa.text("drop table alembic_version"))
            except Exception:  # noqa
                pass

    def drop_all_schemas(self) -> None:
        """Drop all schema in database"""
        with self.sync_engine.begin() as conn:
            conn.execute(sa.text("DROP SCHEMA IF EXISTS public CASCADE"))
            conn.execute(sa.text("CREATE SCHEMA IF NOT EXISTS public"))



import os

from aiohttp import web

from .exc import ImproperlyConfigured

try:
    from . import sentry
except ImportError:  # pragma: no cover
    sentry = None

ERROR_500 = os.environ.get("ERROR_500_MESSSAGE", "Internal Server Error")


def sentry_middleware(app, dsn, env="dev"):
    if not sentry:  # pragma: no cover
        raise ImproperlyConfigured("Sentry middleware requires sentry-sdk")
    sentry.setup(app, dsn, env)


def json_error(status_codes=None):
    status_codes = set(status_codes or (404, 405, 500))
    content_type = "application/json"

    @web.middleware
    async def json_middleware(request, handler):
        try:
            response = await handler(request)
            if response.status not in status_codes:
                return response
            message = response.message
            status = response.status
        except web.HTTPException as ex:
            if ex.status not in status_codes or ex.content_type == content_type:
                raise
            message = ex.reason
            status = ex.status
            if isinstance(message, str):
                message = {"error": message}
        except Exception:
            if 500 in status_codes:
                status = 500
                message = {"error": ERROR_500}
                request.app.logger.exception(ERROR_500)
            else:
                raise
        return web.json_response(message, status=status)

    return json_middleware


# backward compatibility
json404 = json_error((404,))



from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Any, Dict, List, Union

from ..data import fields
from ..data.validate import ValidationErrors
from .channel import logger
from .errors import CONNECTION_ERRORS, CannotPublish, ChannelCallbackError
from .rpc import ws_rpc

if TYPE_CHECKING:  # pragma: no cover
    from .path import WsPathMixin


@dataclass
class PublishSchema:
    data: Union[str, List, Dict]
    channel: str = fields.data_field(
        required=True, description="Channel to publish message"
    )
    event: str = fields.data_field(description="Channel event")


@dataclass
class SubscribeSchema:
    channel: str = fields.data_field(required=True, description="Channel to subscribe")
    event: str = fields.data_field(description="Channel event")


class ChannelCallback:
    """Callback for channels"""

    def __init__(self, ws: "WsPathMixin"):
        self.ws: "WsPathMixin" = ws

    def __repr__(self) -> str:  # pragma: no cover
        return self.ws.socket_id

    def __str__(self) -> str:
        return f"websocket {self.ws.socket_id}"

    async def __call__(self, channel: str, match: str, data: Any) -> None:
        try:
            if hasattr(data, "__call__"):
                data = data()
            await self.ws.write(dict(channel=channel, event=match, data=data))
        except CONNECTION_ERRORS:
            logger.info("lost connection with %s", self)
            await self.ws.close()
            raise ChannelCallbackError
        except Exception:
            logger.exception("Critical exception on connection %s", self)
            await self.ws.close()
            raise ChannelCallbackError


class Publish:
    """Mixin which implements the publish RPC method

    Must be used as mixin of :class:`.WsPathMixin`
    """

    def get_publish_message(self, data: Any) -> Any:
        """Create the publish message from the data payload"""
        return data

    @ws_rpc(body_schema=PublishSchema)
    async def ws_rpc_publish(self, payload):
        """Publish an event on a channel"""
        try:
            event = payload.get("event")
            data = self.get_publish_message(payload.get("data"))
            await self.sockets.publish(payload["channel"], event, data)
            return dict(channel=payload["channel"], event=event, data=data)
        except CannotPublish:
            raise ValidationErrors(dict(channel="Cannot publish to channel"))


class Subscribe:
    """Mixin which implements the subscribe and unsubscribe RPC methods

    Must be used as mixin of :class:`.WsPathMixin`
    """

    @cached_property
    def channel_callback(self) -> ChannelCallback:
        """The callback for :class:`.Channels`"""
        return ChannelCallback(self)

    @ws_rpc(body_schema=SubscribeSchema)
    async def ws_rpc_subscribe(self, payload):
        """Subscribe to an event on a channel"""
        await self.channels.register(
            payload["channel"], payload.get("event"), self.channel_callback
        )
        return dict(subscribed=self.channels.get_subscribed(self.channel_callback))

    @ws_rpc(body_schema=SubscribeSchema)
    async def ws_rpc_unsubscribe(self, payload):
        """Unsubscribe to an event on a channel"""
        await self.channels.unregister(
            payload["channel"], payload.get("event"), self.channel_callback
        )
        return dict(subscribed=self.channels.get_subscribed(self.channel_callback))



import asyncio


class CannotSubscribe(RuntimeError):
    """Raised by a :class:`.ServiceManager`.

    When a :class:`.ServiceManager` is not able to subscribe to a channel
    it should raise this exception
    """


class ChannelCallbackError(RuntimeError):
    """Exception which allow for a clean callback removal"""


class CannotPublish(RuntimeError):
    """Raised when not possible to publish event into channels"""


CONNECTION_ERRORS = (
    asyncio.CancelledError,
    asyncio.TimeoutError,
    RuntimeError,
    ConnectionResetError,
)



import os
from dataclasses import dataclass
from typing import Any, Dict, NoReturn, Optional, cast

from aiohttp import web

from ..types import DataType, QueryType
from ..utils import TypingInfo, as_list, compact
from .dump import dump
from .validate import ErrorType, ValidationErrors, validate

BAD_DATA_MESSAGE = os.getenv("BAD_DATA_MESSAGE", "Invalid data format")


@dataclass
class Operation:
    body_schema: Optional[TypingInfo] = None
    query_schema: Optional[TypingInfo] = None
    response_schema: Optional[TypingInfo] = None
    response: int = 200


class DataView:
    """Utility class for data with a valid :ref:`aio-openapi-schema`"""

    operation: Operation = Operation()

    def cleaned(
        self,
        schema: Any,
        data: QueryType,
        *,
        multiple: bool = False,
        strict: bool = True,
        Error: Optional[type] = None,
    ) -> DataType:
        """Clean data using a given schema

        :param schema: a valid :ref:`aio-openapi-schema` or an the name of an
            attribute in :class:`.Operation`
        :param data: data to validate and clean
        :param multiple: multiple values for a given key are acceptable
        :param strict: all required attributes in schema must be available
        :param Error: optional :class:`.Exception` class
        """
        type_info = self.get_schema(schema)
        validated = validate(type_info, data, strict=strict, multiple=multiple)
        if validated.errors:
            if Error:
                raise Error
            elif schema == "path_schema":
                raise web.HTTPNotFound
            self.raise_validation_error(errors=validated.errors)
        return validated.data

    def dump(self, schema: Any, data: DataType) -> DataType:
        """Dump data using a given a valid :ref:`aio-openapi-schema`,
        if the schema is `None` it returns the same `data` as the input.

        :param schema: a schema or an the name of an attribute in :class:`.Operation`
        :param data: data to clean and dump
        """
        return data if schema is None else dump(self.get_schema(schema), data)

    def get_schema(self, schema: Any = None) -> TypingInfo:
        """Get the :ref:`aio-openapi-schema`. If not found it raises an exception

        :param schema: a schema or an the name of an attribute in :class:`.Operation`
        """
        if isinstance(schema, str):
            Schema = getattr(self.operation, schema, None)
        else:
            Schema = schema
        if Schema is None:
            Schema = getattr(self, str(schema), None)
            if Schema is None:
                raise web.HTTPNotImplemented
        return cast(TypingInfo, TypingInfo.get(Schema))

    def validation_error(
        self, message: str = "", errors: Optional[ErrorType] = None
    ) -> Exception:
        """Create the validation exception used by :meth:`.raise_validation_error`"""
        return ValidationErrors(self.as_errors(message, errors))

    def raise_validation_error(
        self, message: str = "", errors: Optional[ErrorType] = None
    ) -> NoReturn:
        """Raise an :class:`aiohttp.web.HTTPUnprocessableEntity`"""
        raise self.validation_error(message, errors)

    def raise_bad_data(
        self, exc: Optional[Exception] = None, message: str = ""
    ) -> None:
        if not message and exc:
            raise exc from exc
        raise TypeError(message or BAD_DATA_MESSAGE)

    def as_errors(self, message: str = "", errors: Optional[ErrorType] = None) -> Dict:
        if isinstance(errors, str):
            message = cast(str, message or errors)
            errors = None
        return compact(message=message, errors=as_list(errors or ()))



from dataclasses import dataclass
from typing import Dict, NamedTuple, Optional, Type

from multidict import MultiDict
from yarl import URL

from openapi.data.fields import Choice, integer_field, str_field
from openapi.utils import docjoin

from .pagination import (
    DEF_PAGINATION_LIMIT,
    MAX_PAGINATION_LIMIT,
    Pagination,
    PaginationVisitor,
    from_filters_and_dataclass,
)


def offsetPagination(
    *order_by_fields: str,
    default_limit: int = DEF_PAGINATION_LIMIT,
    max_limit: int = MAX_PAGINATION_LIMIT,
) -> Type[Pagination]:
    if len(order_by_fields) == 0:
        raise ValueError("orderable_fields must be specified")

    @dataclass
    class OffsetPagination(Pagination):
        limit: int = integer_field(
            min_value=1,
            max_value=max_limit,
            default=default_limit,
            description="Limit the number of objects returned from the endpoint",
        )
        offset: int = integer_field(
            min_value=0,
            default=0,
            description=(
                "Number of objects to exclude. "
                "Use in conjunction with limit to paginate results"
            ),
        )
        order_by: str = str_field(
            validator=Choice(order_by_fields),
            default=order_by_fields[0],
            description=(
                "Order results by given column (default ascending order). "
                f"Possible values are {docjoin(order_by_fields)}"
            ),
        )

        def apply(self, visitor: PaginationVisitor) -> None:
            visitor.apply_offset_pagination(
                limit=self.limit, offset=self.offset, order_by=self.order_by
            )

        @classmethod
        def create_pagination(cls, data: dict) -> "OffsetPagination":
            return from_filters_and_dataclass(OffsetPagination, data)

        def links(
            self, url: URL, data: list, total: Optional[int] = None
        ) -> Dict[str, str]:
            """Return links for paginated data"""
            return Links(url=url, query=MultiDict(url.query)).links(
                total, self.limit, self.offset
            )

    return OffsetPagination


class Links(NamedTuple):
    url: URL
    query: MultiDict

    def first_link(self, total, limit, offset):
        n = self._count_part(offset, limit, 0)
        if n:
            offset -= n * limit
        if offset > 0:
            return self.link(0, min(limit, offset))

    def prev_link(self, total, limit, offset):
        if offset:
            olimit = min(limit, offset)
            prev_offset = offset - olimit
            return self.link(prev_offset, olimit)

    def next_link(self, total, limit, offset):
        next_offset = offset + limit
        if total > next_offset:
            return self.link(next_offset, limit)

    def last_link(self, total, limit, offset):
        n = self._count_part(total, limit, offset)
        if n > 0:
            return self.link(offset + n * limit, limit)

    def link(self, offset, limit):
        query = self.query.copy()
        query.update({"offset": offset, "limit": limit})
        return self.url.with_query(query)

    def _count_part(self, total, limit, offset):
        n = (total - offset) // limit
        # make sure we account for perfect matching
        if n * limit + offset == total:
            n -= 1
        return max(0, n)

    def links(self, total, limit, offset):
        links = {}
        first = self.first_link(total, limit, offset)
        if first:
            links["first"] = first
            links["prev"] = self.prev_link(total, limit, offset)
        next_ = self.next_link(total, limit, offset)
        if next_:
            links["next"] = next_
            links["last"] = self.last_link(total, limit, offset)
        return links



import hashlib
import logging
import time
from dataclasses import dataclass, field
from typing import Any, Dict

from aiohttp import web

from openapi.ws.channels import Channels

from .. import json
from ..data.validate import ValidationErrors, validated_schema
from ..utils import compact
from .errors import CONNECTION_ERRORS
from .manager import SocketsManager, Websocket

logger = logging.getLogger("openapi.ws")


@dataclass
class RpcProtocol:
    id: str
    method: str
    payload: Dict = field(default_factory=dict)


class ProtocolError(RuntimeError):
    pass


class WsPathMixin(Websocket):
    """Api Path mixin for Websocket RPC protocol"""

    SOCKETS_KEY = "web_sockets"
    """Key in the app where the Web Sockets manager is located"""

    @property
    def sockets(self) -> SocketsManager:
        """Connected websockets"""
        return self.request.app[self.SOCKETS_KEY]

    @property
    def channels(self) -> Channels:
        """Channels for pub/sub"""
        return self.sockets.channels

    async def get(self):
        response = web.WebSocketResponse()
        available = response.can_prepare(self.request)
        if not available:
            raise web.HTTPBadRequest(
                **self.api_response_data(
                    {"message": "Unable to open websocket connection"}
                )
            )
        await response.prepare(self.request)
        self.response = response
        self.started = time.time()
        key = "%s - %s" % (self.request.remote, self.started)
        self.socket_id = hashlib.sha224(key.encode("utf-8")).hexdigest()
        #
        # Add to set of sockets if available
        self.sockets.add(self)
        #
        try:
            async for msg in response:
                if msg.type == web.WSMsgType.TEXT:
                    await self.on_message(msg)
        except CONNECTION_ERRORS:
            logger.info("lost connection with websocket %s", self)
        finally:
            self.sockets.remove(self)
        return response

    def decode_message(self, msg: str) -> Any:
        """Decode JSON string message, override for different protocol"""
        try:
            return json.loads(msg)
        except json.JSONDecodeError:
            raise ProtocolError("JSON string expected") from None

    def encode_message(self, msg: Any) -> str:
        """Encode as JSON string message, override for different protocol"""
        try:
            return json.dumps(msg)
        except TypeError:
            raise ProtocolError("JSON object expected") from None

    async def on_message(self, msg):
        id_ = None
        rpc = None
        try:
            data = self.decode_message(msg.data)
            if not isinstance(data, dict):
                raise ProtocolError(
                    "Malformed message; expected dictionary, "
                    f"got {type(data).__name__}"
                )
            id_ = data.get("id")
            rpc = validated_schema(RpcProtocol, data)
            method = getattr(self, f"ws_rpc_{rpc.method}", None)
            if not method:
                raise ValidationErrors(
                    dict(method=f"{rpc.method} method not available")
                )
            response = await method(rpc.payload or {})
            await self.write(dict(id=rpc.id, method=rpc.method, response=response))
        except ProtocolError as exc:
            logger.error("Protocol error: %s", exc)
            await self.error_message(
                str(exc), id=id_, method=rpc.method if rpc else None
            )
        except ValidationErrors as exc:
            await self.error_message(
                "Invalid RPC parameters",
                errors=exc.errors,
                id=id_,
                method=rpc.method if rpc else None,
            )

    async def error_message(self, message, *, errors=None, **kw):
        error = dict(message=message)
        if errors:
            error["errors"] = errors
        await self.write(compact(error=error, **kw))

    async def write(self, msg: Dict) -> None:
        text = self.encode_message(msg)
        await self.response.send_str(text)

    async def close(self) -> None:
        await self.response.close()
        self.sockets.remove(self)



import logging

import sentry_sdk
from sentry_sdk.integrations.aiohttp import AioHttpIntegration
from sentry_sdk.integrations.logging import LoggingIntegration


def setup(app, dsn, env="dev", level=logging.ERROR, event_level=logging.ERROR):

    sentry_sdk.init(
        dsn=dsn,
        environment=env,
        integrations=[
            LoggingIntegration(
                level=level,  # Capture level and above as breadcrumbs
                event_level=event_level,  # Send event_level and above as events
            ),
            AioHttpIntegration(),
        ],
    )



"""Web socket handler with Publish/Subscribe capabilities

Pub/Sub requires a message broker object in the "broker" app key
"""
from .channel import Channel, Event
from .channels import Channels
from .errors import CannotPublish, CannotSubscribe, ChannelCallbackError
from .manager import SocketsManager, Websocket, WsHandlerType
from .path import WsPathMixin
from .rpc import ws_rpc

__all__ = [
    "WsPathMixin",
    "WsHandlerType",
    "SocketsManager",
    "Websocket",
    "Channels",
    "Channel",
    "Event",
    "CannotPublish",
    "CannotSubscribe",
    "ChannelCallbackError",
    "ws_rpc",
]



from dataclasses import dataclass
from typing import Sequence

from openapi.data.fields import str_field
from openapi.utils import docjoin

from .pagination import from_filters_and_dataclass


class SearchVisitor:
    def apply_search(self, search: str, search_fields: Sequence[str]) -> None:
        raise NotImplementedError


@dataclass
class Search:
    @classmethod
    def create_search(cls, data: dict) -> "Search":
        return cls()

    def apply(self, visitor: SearchVisitor) -> None:
        pass


def searchable(*searchable_fields) -> type:
    """Create a dataclass with `search_fields` class attribute and `search` field.
    The search field is a set of field which can be used for searching and it is used
    internally by the library, while the `search` field is the query string passed
    in the url.

    :param searchable_fields: fields which can be used for searching
    """
    fields = docjoin(searchable_fields)

    @dataclass
    class Searchable(Search):
        search_fields = frozenset(searchable_fields)
        search: str = str_field(
            description=(
                "Search query string. " f"The search is performed on {fields} fields."
            )
        )

        @classmethod
        def create_search(cls, data: dict) -> "Searchable":
            return from_filters_and_dataclass(Searchable, data)

        def apply(self, visitor: SearchVisitor) -> None:
            visitor.apply_search(self.search, self.search_fields)

    return Searchable



import base64
from dataclasses import dataclass
from datetime import date, datetime
from functools import cached_property
from typing import Any, Dict, Optional, Tuple, Type

from dateutil.parser import parse as parse_date
from yarl import URL

from openapi import json
from openapi.data.fields import Choice, integer_field, str_field
from openapi.data.validate import ValidationErrors

from .pagination import (
    DEF_PAGINATION_LIMIT,
    MAX_PAGINATION_LIMIT,
    Pagination,
    PaginationVisitor,
    fields_flip_sign,
    fields_no_sign,
    from_filters_and_dataclass,
)

CursorType = Tuple[Tuple[str, str], ...]


def encode_cursor(data: Tuple[str, ...], previous: bool = False) -> str:
    cursor_bytes = json.dumps((data, previous)).encode("ascii")
    base64_bytes = base64.b64encode(cursor_bytes)
    return base64_bytes.decode("ascii")


def decode_cursor(
    cursor: Optional[str], field_names: Tuple[str]
) -> Tuple[CursorType, bool]:
    try:
        if cursor:
            base64_bytes = cursor.encode("ascii")
            cursor_bytes = base64.b64decode(base64_bytes)
            values, previous = json.loads(cursor_bytes)
            if len(values) == len(field_names):
                return tuple(zip(field_names, values)), previous
            raise ValueError
        return (), False
    except Exception as e:
        raise ValidationErrors("invalid cursor") from e


def cursor_url(url: URL, cursor: str) -> URL:
    query = url.query.copy()
    query.update(_cursor=cursor)
    return url.with_query(query)


def start_values(record: dict, field_names: Tuple[str, ...]) -> Tuple[str, ...]:
    """start values for pagination"""
    return tuple(record[field] for field in field_names)


def cursorPagination(
    *order_by_fields: str,
    default_limit: int = DEF_PAGINATION_LIMIT,
    max_limit: int = MAX_PAGINATION_LIMIT,
) -> Type[Pagination]:

    if len(order_by_fields) == 0:
        raise ValueError("orderable_fields must be specified")

    field_names = fields_no_sign(order_by_fields)

    @dataclass
    class CursorPagination(Pagination):
        limit: int = integer_field(
            min_value=1,
            max_value=max_limit,
            default=default_limit,
            required=False,
            description="Limit the number of objects returned from the endpoint",
        )
        direction: str = str_field(
            validator=Choice(("asc", "desc")),
            required=False,
            default="asc",
            description=("Order results descending or ascending"),
        )
        _cursor: str = ""

        @cached_property
        def cursor_info(self) -> Tuple[CursorType, Tuple[str, ...], bool]:
            order_by = (
                fields_flip_sign(order_by_fields)
                if self.direction == "desc"
                else order_by_fields
            )
            cursor, previous = decode_cursor(self._cursor, order_by)
            return cursor, order_by, previous

        @property
        def previous(self) -> bool:
            return self.cursor_info[2]

        def apply(self, visitor: PaginationVisitor) -> None:
            cursor, order_by, previous = self.cursor_info
            visitor.apply_cursor_pagination(
                cursor,
                self.limit,
                order_by,
                previous=previous,
            )

        @classmethod
        def create_pagination(cls, data: dict) -> "CursorPagination":
            return from_filters_and_dataclass(CursorPagination, data)

        def links(
            self, url: URL, data: list, total: Optional[int] = None
        ) -> Dict[str, str]:
            links = {}
            if self.previous:
                if len(data) > self.limit + 1:
                    links["prev"] = cursor_url(
                        url,
                        encode_cursor(
                            start_values(data[self.limit], field_names), previous=True
                        ),
                    )
                if self._cursor:
                    links["next"] = cursor_url(
                        url,
                        encode_cursor(
                            start_values(data[0], field_names),
                        ),
                    )
            else:
                if len(data) > self.limit:
                    links["next"] = cursor_url(
                        url,
                        encode_cursor(start_values(data[self.limit], field_names)),
                    )
                if self._cursor:
                    links["prev"] = cursor_url(
                        url,
                        encode_cursor(
                            start_values(data[0], field_names),
                            previous=True,
                        ),
                    )
            return links

        def get_data(self, data: list) -> list:
            if self.previous:
                data = list(reversed(data[1:]))
                return data if len(data) <= self.limit else data[1:]
            return data if len(data) <= self.limit else data[: self.limit]

    return CursorPagination


def cursor_to_python(py_type: Type, value: Any) -> Any:
    try:
        if py_type is datetime:
            return parse_date(value)
        elif py_type is date:
            return parse_date(value).date()
        elif py_type is int:
            return int(value)
        else:
            return value
    except Exception as e:
        raise ValidationErrors("invalid cursor") from e



from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Tuple

from openapi.data.validate import ValidationErrors

from .channel import CallbackType, Channel
from .errors import CannotSubscribe

if TYPE_CHECKING:  # pragma: no cover
    from .manager import SocketsManager


class Channels:
    """Manage channels for publish/subscribe"""

    def __init__(self, sockets: "SocketsManager") -> None:
        self.sockets: "SocketsManager" = sockets
        self._channels: Dict[str, Channel] = {}

    @property
    def registered(self) -> Tuple[str, ...]:
        """Registered channels"""
        return tuple(self._channels)

    def __len__(self) -> int:
        return len(self._channels)

    def __contains__(self, channel_name: str) -> bool:
        return channel_name in self._channels

    def __iter__(self) -> Iterator[Channel]:
        return iter(self._channels.values())

    def clear(self) -> None:
        self._channels.clear()

    def get(self, channel_name: str) -> Optional[Channel]:
        return self._channels.get(channel_name)

    def info(self) -> Dict:
        return {channel.name: channel.info() for channel in self}

    async def __call__(self, channel_name: str, message: Dict) -> None:
        """Channel callback"""
        channel = self.get(channel_name)
        if channel:
            closed = await channel(message)
            for websocket in closed:
                for channel_name, channel in tuple(self._channels.items()):
                    channel.remove_callback(websocket)
                    await self._maybe_remove_channel(channel)

    async def register(
        self, channel_name: str, event_name: str, callback: CallbackType
    ) -> Channel:
        """Register a callback

        :param channel_name: name of the channel
        :param event_name: name of the event in the channel or a pattern
        :param callback: the callback to invoke when the `event` on `channel` occurs
        """
        channel = self.get(channel_name)
        if channel is None:
            try:
                await self.sockets.subscribe(channel_name)
            except CannotSubscribe:
                raise ValidationErrors(dict(channel="Invalid channel"))
            else:
                channel = Channel(channel_name)
                self._channels[channel_name] = channel
        event = channel.register(event_name, callback)
        await self.sockets.subscribe_to_event(channel.name, event.name)
        return channel

    async def unregister(
        self, channel_name: str, event: str, callback: CallbackType
    ) -> Optional[Channel]:
        """Safely unregister a callback from the list of event
        callbacks for channel_name
        """
        channel = self.get(channel_name)
        if channel is None:
            raise ValidationErrors(dict(channel="Invalid channel"))
        channel.unregister(event, callback)
        return await self._maybe_remove_channel(channel)

    async def _maybe_remove_channel(self, channel: Channel) -> Channel:
        if not channel:
            await self.sockets.unsubscribe(channel.name)
            self._channels.pop(channel.name)
        return channel

    def get_subscribed(self, callback: CallbackType) -> Dict[str, List[str]]:
        subscribed = {}
        for channel in self:
            events = channel.get_subscribed(callback)
            if events:
                subscribed[channel.name] = events
        return subscribed



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openapi',
 'openapi.data',
 'openapi.db',
 'openapi.db.openapi',
 'openapi.pagination',
 'openapi.spec',
 'openapi.ws']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'SQLAlchemy-Utils>=0.38.2,<0.39.0',
 'SQLAlchemy>=1.4.27,<2.0.0',
 'aiohttp>=3.8.0,<4.0.0',
 'alembic',
 'asyncpg>=0.25.0,<0.26.0',
 'click>=8.0.3,<9.0.0',
 'email-validator>=1.1.3,<2.0.0',
 'httptools>=0.3.0,<0.4.0',
 'psycopg2-binary>=2.9.2,<3.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'pytz>=2022.1,<2023.0',
 'simplejson>=3.17.5,<4.0.0']

extras_require = \
{'dev': ['aiodns>=3.0.0,<4.0.0',
         'PyJWT>=2.3.0,<3.0.0',
         'colorlog>=6.6.0,<7.0.0',
         'phonenumbers>=8.12.37,<9.0.0',
         'cchardet>=2.1.7,<3.0.0'],
 'docs': ['Sphinx>=4.3.0,<5.0.0',
          'sphinx-copybutton>=0.4.0,<0.5.0',
          'sphinx-autodoc-typehints>=1.12.0,<2.0.0',
          'aiohttp-theme>=0.1.6,<0.2.0',
          'recommonmark>=0.7.1,<0.8.0']}

setup_kwargs = {
    'name': 'aio-openapi',
    'version': '2.9.0',
    'description': 'Minimal OpenAPI asynchronous server application',
    'long_description': '# aio-openapi\n\n[![PyPI version](https://badge.fury.io/py/aio-openapi.svg)](https://badge.fury.io/py/aio-openapi)\n[![Python versions](https://img.shields.io/pypi/pyversions/aio-openapi.svg)](https://pypi.org/project/aio-openapi)\n[![Build](https://github.com/quantmind/aio-openapi/workflows/build/badge.svg)](https://github.com/quantmind/aio-openapi/actions?query=workflow%3Abuild)\n[![Coverage Status](https://coveralls.io/repos/github/quantmind/aio-openapi/badge.svg?branch=master)](https://coveralls.io/github/quantmind/aio-openapi?branch=master)\n[![Documentation Status](https://readthedocs.org/projects/aio-openapi/badge/?version=latest)](https://aio-openapi.readthedocs.io/en/latest/?badge=latest)\n[![Downloads](https://img.shields.io/pypi/dd/aio-openapi.svg)](https://pypi.org/project/aio-openapi/)\n\nAsynchronous web middleware for [aiohttp][] and serving Rest APIs with [OpenAPI][] v 3\nspecification and with optional [PostgreSql][] database bindings.\n\n<!-- START doctoc generated TOC please keep comment here to allow auto update -->\n<!-- DON\'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->\n\n**Table of Contents**\n\n- [Installation](#installation)\n- [Development](#development)\n- [Features](#features)\n- [Web App](#web-app)\n- [OpenAPI Documentation](#openapi-documentation)\n- [Database Integration](#database-integration)\n- [Websockets](#websockets)\n  - [RPC protocol](#rpc-protocol)\n  - [Publish/Subscribe](#publishsubscribe)\n- [Environment Variables](#environment-variables)\n\n<!-- END doctoc generated TOC please keep comment here to allow auto update -->\n\n## Installation\n\n```\npip install aio-openapi\n```\n\n## Development\n\nClone the repository and create a virtual environment `venv`.\n\nInstall dependencies by running the install script\n\n```\nmake install\n```\n\nTo run tests\n\n```\nmake test\n```\n\nBy default tests are run against a database with the following connection string `postgresql+asyncpg://postgres:postgres@localhost:5432/openapi`. To use a different DB, create a `.env` file with\na different connection string, for example:\n\n```\nDATASTORE=postgresql+asyncpg://postgres:postgres@localhost:6432/openapi\n```\n\n## Features\n\n- Asynchronous web routes with [aiohttp](https://aiohttp.readthedocs.io/en/stable/)\n- Data validation, serialization and unserialization with python [dataclasses](https://docs.python.org/3/library/dataclasses.html)\n- [OpenApi][] v 3 auto documentation\n- [SqlAlchemy][] expression language\n- Asynchronous DB interaction with [asyncpg][]\n- Migrations with [alembic][]\n- SqlAlchemy tables as python dataclasses\n- Support [click][] command line interface\n- Optional [sentry](https://sentry.io) middleware\n\n## Web App\n\nTo create an openapi RESTful application follow this schema (lets call the file `main.py`)\n\n```python\nfrom openapi.rest import rest\n\ndef create_app():\n    return rest(\n        openapi=dict(\n            title=\'A REST API\',\n            ...\n        ),\n        base_path=\'/v1\',\n        allowed_tags=[...],\n        validate_docs=True,\n        setup_app=setup_app,\n        commands=[...]\n    )\n\n\ndef setup_app(app):\n    app.router.add_routes(...)\n    return app\n\n\nif __name__ == \'__main__\':\n    create_app().main()\n```\n\nThe `create_app` function creates the [aiohttp][] server application by invoking the `rest` function.\nThis function adds the [click][] command in the `cli` mapping entry and add\ndocumentation for routes which support OpenAPI docs.\nThe `setup_app` function is used to actually setup the custom application, usually by adding middleware, routes,\nshutdown callbacks, database integration and so forth.\n\n## OpenAPI Documentation\n\nThe library provide tools for creating OpenAPI v 3 compliant endpoints and\nauto-document them.\n\nAn example from test `tests/example` directory\n\n```python\nfrom typing import List\n\nfrom aiohttp import web\n\nfrom openapi.db.path import SqlApiPath\nfrom openapi.spec import op\n\n\nroutes = web.RouteTableDef()\n\n\n@routes.view(\'/tasks\')\nclass TasksPath(SqlApiPath):\n    """\n    ---\n    summary: Create and query Tasks\n    tags:\n        - name: Task\n          description: Task tag description\n    """\n    table = \'tasks\'\n\n    @op(query_schema=TaskOrderableQuery, response_schema=List[Task])\n    async def get(self) -> web.Response:\n        """\n        ---\n        summary: Retrieve Tasks\n        description: Retrieve a list of Tasks\n        responses:\n            200:\n                description: Authenticated tasks\n        """\n        paginated = await self.get_list()\n        return paginated.json_response()\n\n    @op(response_schema=Task, body_schema=TaskAdd)\n    async def post(self) -> web.Response:\n        """\n        ---\n        summary: Create a Task\n        description: Create a new Task\n        responses:\n            201:\n                description: the task was successfully added\n            422:\n                description: Failed validation\n        """\n        data = await self.create_one()\n        return self.json_response(data, status=201)\n```\n\n## Database Integration\n\nThis library provides integration with [asyncpg][], an high performant asynchronous\nconnector with [PostgreSql][] database.\nTo add the database extension simply use the `get_db` function in the applicatiuon `setup_app` function:\n\n```python\nfrom aiohttp import web\n\nfrom openapi.db import get_db\n\ndef setup_app(app: web.Application) -> None:\n    db = get_db(app)\n    meta = db.metadata\n\n```\n\nThis will enable database connection and command line tools (most of them from [alembic][]):\n\n```\npython main.py db --help\n```\n\nThe database container is available at the `db` app key:\n\n```python\napp[\'db\']\n```\n\n## Environment Variables\n\nSeveral environment variables are used by the library to support testing and deployment.\n\n- `DATASTORE`: PostgreSql connection string (same as [SqlAlchemy][] syntax)\n- `DBPOOL_MIN_SIZE`: minimum size of database connection pool (default is 10)\n- `DBPOOL_MAX_SIZE`: maximum size of database connection pool (default is 10)\n\n[aiohttp]: https://aiohttp.readthedocs.io/en/stable/\n[openapi]: https://www.openapis.org/\n[postgresql]: https://www.postgresql.org/\n[sqlalchemy]: https://www.sqlalchemy.org/\n[click]: https://github.com/pallets/click\n[alembic]: http://alembic.zzzcomputing.com/en/latest/\n[asyncpg]: https://github.com/MagicStack/asyncpg\n',
    'author': 'Luca',
    'author_email': 'luca@quantmind.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/quantmind/aio-openapi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



def redis_to_py_pattern(pattern):
    return "".join(_redis_to_py_pattern(pattern))


def _redis_to_py_pattern(pattern):
    clear, esc = False, False
    s, q, op, cp, e = "*", "?", "[", "]", "\\"

    for v in pattern:
        if v == s and not esc:
            yield "(.*)"
        elif v == q and not esc:
            yield "."
        elif v == op and not esc:
            esc = True
            yield v
        elif v == cp and esc:
            esc = False
            yield v
        elif v == e:
            clear, esc = True
            yield v
        elif clear:
            clear, esc = False, False
            yield v
        else:
            yield v
    yield "$"



import uuid

import sqlalchemy as sa
from sqlalchemy_utils import UUIDType


def UUIDColumn(name, primary_key=True, nullable=False, make_default=False, **kw):
    if primary_key and nullable:
        raise RuntimeError("Primary key must be NOT NULL")
    if make_default:
        kw.setdefault("default", uuid.uuid4)
    return sa.Column(name, UUIDType(), primary_key=primary_key, nullable=nullable, **kw)



from typing import Dict, List

from aiohttp import web

from .path import full_url


def default_server(request: web.Request) -> Dict[str, str]:
    app = request.app
    url = full_url(request)
    url = url.with_path(app["cli"].base_path)
    return dict(url=str(url), description="Api server")


def server_urls(request: web.Request, paths: List[str]) -> List[str]:
    base_path = request.app["cli"].base_path
    n = len(base_path)
    spec = request.app.get("spec")
    server = spec.servers[0] if spec and spec.servers else default_server(request)
    base_url = server["url"]
    return [f"{base_url}{p[n:]}" for p in paths]



from datetime import datetime
from enum import Enum
from functools import partial
from uuid import UUID

import simplejson
from simplejson.errors import JSONDecodeError


def encoder(obj):
    if isinstance(obj, UUID):
        return obj.hex
    if isinstance(obj, datetime):
        return obj.isoformat()
    if isinstance(obj, Enum):
        return obj.name
    raise TypeError


loads = partial(simplejson.loads, use_decimal=True)
dumps = partial(
    simplejson.dumps, use_decimal=True, default=encoder, iterable_as_array=True
)

__all__ = ["loads", "dumps", "JSONDecodeError"]



from dataclasses import Field, make_dataclass
from datetime import date, datetime
from decimal import Decimal
from functools import partial
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

import sqlalchemy as sa
from sqlalchemy_utils import UUIDType

from . import fields

ConverterType = Callable[[sa.Column, bool, bool, Sequence[str]], Tuple[Type, Field]]
CONVERTERS: Dict[str, ConverterType] = {}


def dataclass_from_table(
    name: str,
    table: sa.Table,
    *,
    exclude: Optional[Sequence[str]] = None,
    include: Optional[Sequence[str]] = None,
    default: Union[bool, Sequence[str]] = False,
    required: Union[bool, Sequence[str]] = False,
    ops: Optional[Dict[str, Sequence[str]]] = None,
) -> type:
    """Create a dataclass from an :class:`sqlalchemy.schema.Table`

    :param name: dataclass name
    :param table: sqlalchemy table
    :param exclude: fields to exclude from the dataclass
    :param include: fields to include in the dataclass
    :param default: use columns defaults in the dataclass
    :param required: set non nullable columns without a default as
        required fields in the dataclass
    :param ops: additional operation for fields
    """
    columns = []
    includes = set(include or table.columns.keys()) - set(exclude or ())
    defaults = column_info(includes, default)
    requireds = column_info(includes, required)
    column_ops = cast(Dict[str, Sequence[str]], ops or {})
    for col in table.columns:
        if col.name not in includes:
            continue
        ctype = type(col.type)
        converter = CONVERTERS.get(ctype)
        if not converter:  # pragma:   no cover
            raise NotImplementedError(f"Cannot convert column {col.name}: {ctype}")
        required = col.name in requireds
        use_default = col.name in defaults
        field = (
            col.name,
            *converter(col, required, use_default, column_ops.get(col.name, ())),
        )
        columns.append(field)
    return make_dataclass(name, columns)


def column_info(columns: Set[str], value: Union[bool, Sequence[str]]) -> Set[str]:
    if value is False:
        return set()
    elif value is True:
        return columns.copy()
    else:
        return set(value if value is not None else columns)


def converter(*types):
    def _(f):
        for type_ in types:
            CONVERTERS[type_] = f
        return f

    return _


@converter(sa.Boolean)
def bl(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.bool_field)
    return (bool, data_field(**info(col, required, use_default, ops)))


@converter(sa.Integer)
def integer(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.number_field)
    return (int, data_field(precision=0, **info(col, required, use_default, ops)))


@converter(sa.Numeric)
def number(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.decimal_field)
    return (
        Decimal,
        data_field(precision=col.type.scale, **info(col, required, use_default, ops)),
    )


@converter(sa.String, sa.Text, sa.CHAR, sa.VARCHAR)
def string(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.str_field)
    return (
        str,
        data_field(
            max_length=col.type.length or 0, **info(col, required, use_default, ops)
        ),
    )


@converter(sa.DateTime)
def dt_ti(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.date_time_field)
    return (
        datetime,
        data_field(timezone=col.type.timezone, **info(col, required, use_default, ops)),
    )


@converter(sa.Date)
def dt(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.date_field)
    return (date, data_field(**info(col, required, use_default, ops)))


@converter(sa.Enum)
def en(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.enum_field)
    return (
        col.type.enum_class,
        data_field(col.type.enum_class, **info(col, required, use_default, ops)),
    )


@converter(sa.JSON)
def js(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.json_field)
    val = None
    if col.default:
        arg = col.default.arg
        val = arg() if col.default.is_callable else arg
    return (
        JsonTypes.get(type(val), Dict),
        data_field(**info(col, required, use_default, ops)),
    )


@converter(UUIDType)
def uuid(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data_field = col.info.get("data_field", fields.uuid_field)
    return (str, data_field(**info(col, required, use_default, ops)))


def info(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> Tuple[Type, Field]:
    data = dict(ops=ops)
    if use_default:
        default = col.default.arg if col.default is not None else None
        if callable(default):
            data.update(default_factory=partial(default, None))
            required = False
        elif isinstance(default, (list, dict, set)):
            data.update(default_factory=lambda: default.copy())
            required = False
        else:
            data.update(default=default)
            if required and (col.nullable or default is not None):
                required = False
    elif required and col.nullable:
        required = False
    data.update(required=required)
    if col.doc:
        data.update(description=col.doc)
    data.update(col.info)
    data.pop("data_field", None)
    return data


JsonTypes = {list: List, dict: Dict}



from typing import Optional, Type, TypeVar

from ..utils import TypingInfo
from .pagination import Pagination
from .search import Search

T = TypeVar("T")

CREATE_MAP = {Pagination: "create_pagination", Search: "create_search"}


def create_dataclass(
    type_info: Optional[TypingInfo], data: dict, DataClass: Type[T]
) -> T:
    if type_info is None:
        return DataClass()
    if type_info.is_dataclass and issubclass(type_info.element, DataClass):
        method_name = CREATE_MAP.get(DataClass)
        if method_name:
            return getattr(type_info.element, method_name)(data)
    return DataClass()
